"""
 @file
 @brief This file is for legacy support of OpenShot 1.x project files
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """


class sequence:
    """A sequence contains tracks and clips that make up a scene (aka sequence).  Currently, Openshot
    only contains a single sequence, but soon it will have the ability to create many sequences."""

    # ----------------------------------------------------------------------
    def __init__(self, seq_name, project):
        """Constructor"""

        # init variables for sequence
        self.name = seq_name
        self.length = 600.0  # length in seconds of sequence.  This controls how wide to render the tracks.
        self.project = project  # reference to current project
        self.scale = 8.0  # this represents how many seconds per tick mark
        self.tick_pixels = 100  # number of pixels between each tick mark
        self.play_head_position = 0.0  # position of the play head in seconds

        # init the tracks on the sequence
        self.tracks = []

        # init markers
        self.markers = []

        # reference to play_head goocanvas group
        self.play_head = None
        self.ruler_time = None
        self.play_head_line = None
        self.enable_animated_playhead = True



""" 
 @file
 @brief This file contains helper functions for Qt types (string to base64)
 @author Noah Figg <eggmunkee@hotmail.com>
 @author Jonathan Thomas <jonathan@openshot.org>
 @author Olivier Girard <eolinwen@gmail.com>
 
 @section LICENSE
 
 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.
 
 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from PyQt5.QtCore import QByteArray


# Utility functions for handling qt types

# QByteArray helpers
def str_to_bytes(string):
    """ This is required to save Qt byte arrays into a base64 string (to save screen preferences) """
    return QByteArray.fromBase64(string.encode("utf-8"))


def bytes_to_str(bytes):
    """ This is required to load base64 Qt byte array strings into a Qt byte array (to load screen preferences) """
    return bytes.toBase64().data().decode("utf-8")


def model_index_sibling_at_column(index, column):
    """Return a sibling index for a column across old/new Qt bindings."""
    sibling_at_column = getattr(index, "siblingAtColumn", None)
    if callable(sibling_at_column):
        return sibling_at_column(column)
    return index.sibling(index.row(), column)


def font_metrics_horizontal_advance(metrics, text):
    """Measure text width across old/new Qt bindings."""
    horizontal_advance = getattr(metrics, "horizontalAdvance", None)
    if callable(horizontal_advance):
        return horizontal_advance(text)
    return metrics.width(text)



"""
 @file
 @brief This file loads the profile editor (duplicate and edit profiles)
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2024 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os

import openshot
from PyQt5.QtWidgets import QDialog, QMessageBox
from classes import ui_util, info, tabstops
from classes.app import get_app
from classes.logger import log

MAX_FPS_SPINBOX_VALUE = 2147483647


class EditProfileDialog(QDialog):
    """ Edit Profile Dialog """

    # Path to ui file
    ui_path = os.path.join(info.PATH, 'windows', 'ui', 'profile-edit.ui')

    def __init__(self, profile, duplicate):
        super(EditProfileDialog, self).__init__()

        # Make copy of profile
        self.original_profile = profile.Json()
        self.profile = profile
        self.duplicate = duplicate
        if duplicate:
            # Clear reference to original profile
            self.profile = openshot.Profile()
            self.profile.SetJson(self.original_profile)

        # Save initial file path (if editing, this needs to be removed)
        self.existing_path = None
        if hasattr(profile, "path"):
            self.existing_path = profile.path

        # Load UI from designer & init
        ui_util.load_ui(self, self.ui_path)
        ui_util.init_ui(self)

        # Populate fields from profile
        self.initialize()

        tabstops.apply_auto_tab_order_later(self)

    def initialize(self):
        """Initialize the form fields with data from the profile."""
        # get translations
        _ = get_app()._tr

        # Update windows title
        if self.duplicate:
            self.setWindowTitle(_('Create Profile'))
        else:
            self.setWindowTitle(_('Edit Profile'))

        # Add options to cboInterlaced dropdown
        self.cboInterlaced.addItem(_('Yes'))
        self.cboInterlaced.addItem(_('No'))

        # Add options to cboSpherical dropdown
        self.cboSpherical.addItem(_('Yes'))
        self.cboSpherical.addItem(_('No'))

        self.txtProfileName.setText(self.profile.info.description)
        self.txtWidth.setValue(self.profile.info.width)
        self.txtHeight.setValue(self.profile.info.height)
        self.txtAspectRatioNum.setValue(self.profile.info.display_ratio.num)
        self.txtAspectRatioDen.setValue(self.profile.info.display_ratio.den)
        self.txtPixelRatioNum.setValue(self.profile.info.pixel_ratio.num)
        self.txtPixelRatioDen.setValue(self.profile.info.pixel_ratio.den)
        self.txtFrameRateNum.setMaximum(MAX_FPS_SPINBOX_VALUE)
        self.txtFrameRateDen.setMaximum(MAX_FPS_SPINBOX_VALUE)
        self.txtFrameRateNum.setValue(self.profile.info.fps.num)
        self.txtFrameRateDen.setValue(self.profile.info.fps.den)
        self.update_frame_rate_display()
        self.cboInterlaced.setCurrentText(_('Yes') if self.profile.info.interlaced_frame == 1 else _('No'))
        self.cboSpherical.setCurrentText(_('Yes') if self.profile.info.spherical else _('No'))
        if not self.duplicate:
            self.update_profile_path()

        # Connect all signals
        self.connect_signals()

    def connect_signals(self):
        """Connect input fields to update profile on change."""
        self.txtProfileName.textChanged.connect(self.update_profile_description)
        self.txtWidth.valueChanged.connect(self.update_profile_width)
        self.txtHeight.valueChanged.connect(self.update_profile_height)
        self.txtPixelRatioNum.valueChanged.connect(self.update_profile_pixel_ratio)
        self.txtPixelRatioDen.valueChanged.connect(self.update_profile_pixel_ratio)
        self.txtFrameRateNum.valueChanged.connect(self.update_profile_frame_rate_num)
        self.txtFrameRateDen.valueChanged.connect(self.update_profile_frame_rate_den)
        self.txtFrameRateNum.valueChanged.connect(self.update_frame_rate_display)
        self.txtFrameRateDen.valueChanged.connect(self.update_frame_rate_display)
        self.cboInterlaced.currentTextChanged.connect(self.update_profile_interlaced)
        self.cboSpherical.currentTextChanged.connect(self.update_profile_spherical)
        self.txtProfileName.setFocus()

    # Handlers for updating profile
    def update_profile_description(self, value):
        self.profile.info.description = value
        self.update_profile_path()

    def update_profile_width(self, value):
        self.profile.info.width = value
        self.update_display_aspect_ratio()
        self.update_profile_path()

    def update_profile_height(self, value):
        self.profile.info.height = value
        self.update_display_aspect_ratio()
        self.update_profile_path()

    def update_profile_pixel_ratio(self):
        num = self.txtPixelRatioNum.value()
        den = self.txtPixelRatioDen.value()
        self.profile.info.pixel_ratio.num = num
        self.profile.info.pixel_ratio.den = den
        self.update_display_aspect_ratio()
        self.update_profile_path()

    def update_display_aspect_ratio(self):
        """Update display aspect ratio based on width, height, and pixel ratio."""
        width = self.profile.info.width
        height = self.profile.info.height
        pixel_ratio_num = self.profile.info.pixel_ratio.num
        pixel_ratio_den = self.profile.info.pixel_ratio.den

        if height != 0 and pixel_ratio_den != 0:
            self.profile.info.display_ratio.num = width * pixel_ratio_num
            self.profile.info.display_ratio.den = height * pixel_ratio_den
            self.profile.info.display_ratio.Reduce()

        # Update the aspect ratio fields in the UI
        self.txtAspectRatioNum.setValue(self.profile.info.display_ratio.num)
        self.txtAspectRatioDen.setValue(self.profile.info.display_ratio.den)

    def update_profile_frame_rate_num(self, value):
        self.profile.info.fps.num = value
        self.update_profile_path()

    def update_profile_frame_rate_den(self, value):
        self.profile.info.fps.den = value
        self.update_profile_path()

    def update_frame_rate_display(self):
        fps_den = self.txtFrameRateDen.value() or 1
        fps_float = self.txtFrameRateNum.value() / fps_den
        self.lblFrameRateValueDisplay.setText(f"= {fps_float:.2f}")

    def update_profile_interlaced(self, value):
        # get translations
        _ = get_app()._tr

        self.profile.info.interlaced_frame = True if value == _('Yes') else False
        self.update_profile_path()

    def update_profile_spherical(self, value):
        # get translations
        _ = get_app()._tr

        self.profile.info.spherical = True if value == _('Yes') else False
        self.update_profile_path()

    def update_profile_path(self):
        if self.existing_path and os.path.exists(self.existing_path) and not self.duplicate:
            profiles_path = self.existing_path
        else:
            profiles_path = os.path.join(info.USER_PROFILES_PATH, self.profile.Key())

        profile_suffix = 1
        while self.duplicate and os.path.exists(profiles_path):
            # Add suffix if 'duplicate' mode - and existing file found
            profiles_path = f"{os.path.join(info.USER_PROFILES_PATH, self.profile.Key())}-{profile_suffix}"
            profile_suffix += 1
        self.lblFilePathValue.setText(profiles_path)

    def accept(self):
        """Save the profile to a file when the user accepts the dialog."""
        # get translations
        _ = get_app()._tr

        # Ensure a concrete output path exists even if the placeholder is still shown.
        self.update_profile_path()

        # Prevent saving with no description
        error_title = _("Profile Error")
        error_message = _("Please enter a <b>unique</b> description for this profile.")
        if not self.profile.info.description.strip():
            QMessageBox.warning(self, error_title, error_message)
            return

        # Verify description is unique
        for profile_folder in [info.USER_PROFILES_PATH, info.PROFILES_PATH]:
            for file in reversed(sorted(os.listdir(profile_folder))):
                profile_verify_path = os.path.join(profile_folder, file)
                if os.path.isdir(profile_verify_path) or profile_verify_path == self.lblFilePathValue.text():
                    continue
                try:
                    # Load Profile
                    p = openshot.Profile(profile_verify_path)
                    if p.info.description.strip() == self.profile.info.description.strip():
                        QMessageBox.warning(self, error_title, error_message)
                        return
                except RuntimeError as e:
                    log.warning("Failed to parse file '%s' as a profile: %s" % (profile_verify_path, e))

        # Save the profile data as a text file in the user profiles folder
        profile_path = self.lblFilePathValue.text()
        log.info(f"Saving custom profile: {profile_path}")
        self.profile.Save(profile_path)
        self.profile.user_created = True
        self.profile.path = profile_path

        # Accept the dialog
        super(EditProfileDialog, self).accept()

    def reject(self):
        """Close the dialog without saving changes."""
        # restore original profile
        self.profile.SetJson(self.original_profile)

        super(EditProfileDialog, self).reject()



"""
 @file
 @brief This file converts a float time value (i.e. seconds) to parts of time
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import math


def padNumber(value, pad_length):
    """Pad number a specific # of characters"""
    format_mask = '%%0%sd' % pad_length
    return format_mask % value


def secondsToTime(secs, fps_num=30, fps_den=1):
    """Convert # of seconds (float) to parts of time (dict)"""
    milliseconds = secs * 1000
    sec = math.floor(milliseconds / 1000)
    milli = milliseconds % 1000
    minute = math.floor(sec / 60)
    sec = sec % 60
    hour = math.floor(minute / 60)
    minute = minute % 60
    day = math.floor(hour / 24)
    hour = hour % 24
    week = math.floor(day / 7)
    day = day % 7

    fps_float = (float(fps_num) / float(fps_den)) if fps_den else 0.0
    frame = int(round((milli / 1000.0) * fps_float))
    if fps_float > 0:
        frame = max(0, min(frame, int(fps_float) - 1))
    return {"week": padNumber(week, 2), "day": padNumber(day, 2), "hour": padNumber(hour, 2),
            "min": padNumber(minute, 2), "sec": padNumber(sec, 2), "milli": padNumber(milli, 3),
            "frame": padNumber(frame, 2)}

def timecodeToSeconds(time_code="00:00:00:00", fps_num=30, fps_den=1):
    """Convert time code to seconds (float)"""
    fps_float = float(fps_num / fps_den)

    seconds = 0.0
    time_parts = time_code.split(":")
    if len(time_parts) == 4:
        hours = float(time_parts[0])
        mins = float(time_parts[1])
        secs = float(time_parts[2])
        frames = float(time_parts[3])
        seconds = (hours * 60 * 60) + (mins * 60) + secs + (frames / fps_float)
    return seconds

def secondsToTimecode(time_in_seconds=0.0, fps_num=30, fps_den=1, use_milliseconds=False):
    """Return a formatted time code HH:MM:SS:FRAME"""
    if use_milliseconds:
        return "%(hour)s:%(min)s:%(sec)s:%(milli)s" % secondsToTime(time_in_seconds, fps_num, fps_den)
    return "%(hour)s:%(min)s:%(sec)s:%(frame)s" % secondsToTime(time_in_seconds, fps_num, fps_den)



#!/usr/bin/python3
"""
 @file
 @brief This file updates the OpenShot.POT (language translation template) by scanning all source files.
 @author Jonathan Thomas <jonathan@openshot.org>

 This file helps you generate the POT file that contains all of the translatable
 strings / text in OpenShot.  Because some of our text is in custom XML files,
 the xgettext command can't correctly generate the POT file.  Thus... the
 existence of this file. =)

 Command to create the individual language PO files (Ascii files)
		$ msginit --input=OpenShot.pot --locale=fr_FR
		$ msginit --input=OpenShot.pot --locale=es

 Command to update the PO files (if text is added or changed)
		$ msgmerge en_US.po OpenShot.pot -U
		$ msgmerge es.po OpenShot.pot -U

 Command to compile the Ascii PO files into binary MO files
		$ msgfmt en_US.po --output-file=en_US/LC_MESSAGES/OpenShot.mo
		$ msgfmt es.po --output-file=es/LC_MESSAGES/OpenShot.mo

 Command to compile all PO files in a folder
		$ find -iname "*.po" -exec msgfmt {} -o {}.mo \\;

 Command to combine the 2 pot files into 1 file
       $ msgcat ~/openshot/locale/OpenShot/OpenShot_source.pot ~/openshot/openshot/locale/OpenShot/OpenShot_glade.pot -o ~/openshot/main/locale/OpenShot/OpenShot.pot

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import shutil
import datetime
import os
import subprocess
import sys
import re
import json

# Try to get the security-patched XML functions from defusedxml
try:
  from defusedxml import minidom as xml
except ImportError:
  from xml.dom import minidom as xml

import openshot

# Get the absolute path of this project
path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if path not in sys.path:
    sys.path.append(path)

import classes.info as info
from classes.logger import log
from classes.effect_init import effect_options

# get the path of the main OpenShot folder
language_folder_path = os.path.dirname(os.path.abspath(__file__))
openshot_path = os.path.dirname(language_folder_path)
effects_path = os.path.join(openshot_path, 'effects')
blender_path = os.path.join(openshot_path, 'blender')
transitions_path = os.path.join(openshot_path, 'transitions')
titles_path = os.path.join(openshot_path, 'titles')
export_path = os.path.join(openshot_path, 'presets')
windows_ui_path = os.path.join(openshot_path, 'windows', 'ui')

log.info("-----------------------------------------------------")
log.info(" Creating temp POT files")
log.info("-----------------------------------------------------")

# create empty temp POT files
temp_files = ['OpenShot_source.pot', 'OpenShot_glade.pot', 'OpenShot_effects.pot', 'OpenShot_export.pot',
              'OpenShot_transitions.pot', 'OpenShot_QtUi.pot']
for temp_file_name in temp_files:
    temp_file_path = os.path.join(language_folder_path, temp_file_name)
    if os.path.exists(temp_file_path):
        os.remove(temp_file_path)
    f = open(temp_file_path, "w")
    f.close()

log.info("-----------------------------------------------------")
log.info(" Using xgettext to generate .py POT files")
log.info("-----------------------------------------------------")

# Generate POT for Source Code strings (i.e. strings marked with a _("translate me"))
subprocess.call(r'find %s -iname "*.py" -exec xgettext -j -o %s --keyword=_ {} \;' % (
    openshot_path, os.path.join(language_folder_path, 'OpenShot_source.pot')), shell=True)

log.info("-----------------------------------------------------")
log.info(" Using Qt's lupdate to generate .ui POT files")
log.info("-----------------------------------------------------")

# Generate POT for Qt *.ui files (which require the lupdate command, and ts2po command)
os.chdir(windows_ui_path)
subprocess.call('lupdate *.ui -ts %s' % (os.path.join(language_folder_path, 'OpenShot_QtUi.ts')), shell=True)
subprocess.call('lupdate *.ui -ts %s' % (os.path.join(language_folder_path, 'OpenShot_QtUi.pot')), shell=True)
os.chdir(language_folder_path)

# Rewrite the UI POT, removing msgctxt
output = open(os.path.join(language_folder_path, "clean.po"), 'w')
for line in open(os.path.join(language_folder_path, 'OpenShot_QtUi.pot'), 'r'):
    if not line.startswith('msgctxt'):
        output.write(line)
# Overwrite original PO file
output.close()
shutil.copy(os.path.join(language_folder_path, "clean.po"), os.path.join(language_folder_path, 'OpenShot_QtUi.pot'))
os.remove(os.path.join(language_folder_path, "clean.po"))

# Remove duplicates (if any found)
subprocess.call('msguniq %s --use-first -o %s' % (os.path.join(language_folder_path, 'OpenShot_QtUi.pot'),
                                                  os.path.join(language_folder_path, 'clean.po')), shell=True)
shutil.copy(os.path.join(language_folder_path, "clean.po"), os.path.join(language_folder_path, 'OpenShot_QtUi.pot'))
os.remove(os.path.join(language_folder_path, "clean.po"))


log.info("-----------------------------------------------------")
log.info(" Updating auto created POT files to set CharSet")
log.info("-----------------------------------------------------")

temp_files = ['OpenShot_source.pot', 'OpenShot_glade.pot']
for temp_file in temp_files:
    # get the entire text
    f = open(os.path.join(language_folder_path, temp_file), "r")
    # read entire text of file
    entire_source = f.read()
    f.close()

    # replace charset
    entire_source = entire_source.replace("charset=CHARSET", "charset=UTF-8")

    # Create Updated POT Output File
    if os.path.exists(os.path.join(language_folder_path, temp_file)):
        os.remove(os.path.join(language_folder_path, temp_file))
    f = open(os.path.join(language_folder_path, temp_file), "w")
    f.write(entire_source)
    f.close()

log.info("-----------------------------------------------------")
log.info(" Scanning effects & resources used by OpenShot")
log.info("-----------------------------------------------------")

props = json.loads(openshot.Clip().PropertiesJSON(1))

# Loop through props
effects_text = {}
for key in props.keys():
    property = props[key]
    if "name" in property:
        effects_text[property["name"]] = "libopenshot (Clip Properties)"
    if "choices" in property:
        for choice in property["choices"]:
            effects_text[choice["name"]] = "libopenshot (Clip Properties)"

# Loop through each libopenshot effect
objects = json.loads(openshot.EffectInfo.Json())
for object in objects:
    class_name = object.get("class_name")
    props = json.loads(openshot.EffectInfo().CreateEffect(class_name).PropertiesJSON(1))

    # Loop through props
    for key in props.keys():
        property = props[key]
        if key == "objects":
            continue # Skip tracker / object detection property
        if "name" in property:
            effects_text[property["name"]] = "libopenshot (Effect Properties)"
        if "choices" in object:
            for choice in property["choices"]:
                effects_text[choice["name"]] = "libopenshot (Effect Properties)"

# Append Effect Init Data
# Loop through props
for effect in effect_options:
    for param in effect_options[effect]:
        if "title" in param:
            effects_text[param["title"]] = "effect_init (Effect parameter for %s)" % effect
        if "values" in param:
            for value in param["values"]:
                effects_text[value["name"]] = "effect_init (Effect parameter for %s)" % effect

# Append Effect Meta Data
e = openshot.EffectInfo()
props = json.loads(e.Json())

# Loop through props
for effect in props:
    if "name" in effect:
        effects_text[effect["name"]] = "libopenshot (Effect Metadata)"
    if "description" in effect:
        effects_text[effect["description"]] = "libopenshot (Effect Metadata)"

# Append LUT category and file names (for ColorMap effect)
for folder in os.listdir(info.COLORS_PATH):
    category_name = folder.replace("_", " ").title()
    folder_path = os.path.join(info.COLORS_PATH, folder)
    if os.path.isdir(folder_path):
        for filename in os.listdir(folder_path):
            basename, extension = os.path.splitext(filename)
            if filename.endswith(".cube"):
                lut_name = basename.replace("_", " ").title()
                effects_text[category_name] = "ColorMap effect lookup (Category)"
                effects_text[lut_name] = "ColorMap effect lookup (Name)"

# Append Emoji Data
emoji_text = { "translator-credits": "Translator credits to be translated by LaunchPad" }
emoji_metadata_path = os.path.join(info.PATH, "emojis", "data", "openmoji-optimized.json")
emoji_ignore_keys = ("Keyboard", "Sunset", "Key", "Right arrow", "Left arrow", "Bubbles",
                     "Twitter", "Instagram", "Scale", "Simple", "Close", "Forward", "Copy",
                     "Filter", "Details", "Duplicate", "Edit", "Delete")
with open(emoji_metadata_path, 'r', encoding="utf-8") as f:
    emoji_metadata = json.load(f)

    # Loop through props
    for filename, emoji in emoji_metadata.items():
        emoji_name = emoji["annotation"].capitalize()
        emoji_group = emoji["group"].split('-')[0].capitalize()
        if "annotation" in emoji and emoji_name not in emoji_ignore_keys:
            emoji_text[emoji_name] = "Emoji Metadata (Displayed Name)"
        if "group" in emoji and emoji_group not in effects_text and emoji_group not in emoji_ignore_keys:
            emoji_text[emoji_group] = "Emoji Metadata (Group Filter name)"

# Loop through the Blender XML
blender_text = { "translator-credits": "Translator credits to be translated by LaunchPad" }
blender_ignore_keys = ("Title", "Alpha", "Blur", "Font Name", "Yes", "No", "On", "Off", "Default")
for file in os.listdir(blender_path):
    if os.path.isfile(os.path.join(blender_path, file)):
        # load xml effect file
        full_file_path = os.path.join(blender_path, file)
        xmldoc = xml.parse(os.path.join(blender_path, file))

        # add text to list
        translation_key = xmldoc.getElementsByTagName("title")[0].childNodes[0].data
        if translation_key not in blender_ignore_keys:
            blender_text[translation_key] = full_file_path

        # get params
        params = xmldoc.getElementsByTagName("param")

        # Loop through params
        for param in params:
            if param.attributes["title"]:
                translation_key = param.attributes["title"].value
                if translation_key not in blender_ignore_keys:
                    blender_text[param.attributes["title"].value] = full_file_path

                    # Loop through child nodes of each param
                    for child in param.childNodes:
                        if child.nodeName == "values":
                            for value in child.getElementsByTagName("value"):
                                if value.hasAttribute("name"):
                                    value_name = value.getAttribute("name")
                                    if value_name not in blender_ignore_keys:
                                        blender_text[value_name] = full_file_path

# Loop through the Export Settings XML
export_text = {}
for file in os.listdir(export_path):
    if os.path.isfile(os.path.join(export_path, file)):
        # load xml export file
        full_file_path = os.path.join(export_path, file)
        xmldoc = xml.parse(os.path.join(export_path, file))

        # add text to list
        export_text[xmldoc.getElementsByTagName("type")[0].childNodes[0].data] = full_file_path
        export_text[xmldoc.getElementsByTagName("title")[0].childNodes[0].data] = full_file_path

# Loop through Settings
settings_file = open(os.path.join(info.PATH, 'settings', '_default.settings'), 'r').read()
settings = json.loads(settings_file)
category_names = []
for setting in settings:
    if "type" in setting and setting["type"] != "hidden":
        # Add visible settings
        export_text[setting["title"]] = "Settings for %s" % setting["setting"]
    if "type" in setting and setting["type"] != "hidden":
        # Add visible category names
        if setting["category"] not in category_names:
            export_text[setting["category"]] = "Settings Category for %s" % setting["category"]
            category_names.append(setting["category"])
        if "translate_values" in setting and setting.get("translate_values"):
            # Add translatable dropdown keys
            for value in setting.get("values", []):
                export_text[value["name"]] = "Settings for %s" % setting["setting"]

# Include UI Theme Names (for translation)
from themes.manager import ThemeName
for theme_name in ThemeName.get_sorted_theme_names():
    export_text[theme_name] = "User-Interface Theme Name"

# Loop through transitions and add to POT file
transitions_text = { "translator-credits": "Translator credits to be translated by LaunchPad" }
transitions_ignore_keys = ("Common", "Fade")
for file in os.listdir(transitions_path):
    # load xml export file
    full_file_path = os.path.join(transitions_path, file)
    (fileBaseName, fileExtension) = os.path.splitext(file)

    # get transition name
    name = fileBaseName.replace("_", " ").capitalize()

    # add text to list
    if name not in transitions_ignore_keys:
        transitions_text[name] = full_file_path

    # Look in sub-folders
    for sub_file in os.listdir(full_file_path):
        # load xml export file
        full_subfile_path = os.path.join(full_file_path, sub_file)
        fileBaseName = os.path.splitext(sub_file)[0]

        # split the name into parts (looking for a number)
        suffix_number = None
        name_parts = fileBaseName.split("_")
        if name_parts[-1].isdigit():
            suffix_number = name_parts[-1]

        # get transition name
        name = fileBaseName.replace("_", " ").capitalize()

        # replace suffix number with placeholder (if any)
        if suffix_number:
            name = name.replace(suffix_number, "%s")

        # add text to list
        if name not in transitions_ignore_keys:
            transitions_text[name] = full_subfile_path

# Loop through titles and add to POT file
for sub_file in os.listdir(titles_path):
    # load xml export file
    full_subfile_path = os.path.join(titles_path, sub_file)
    fileBaseName = os.path.splitext(sub_file)[0]

    # split the name into parts (looking for a number)
    suffix_number = None
    name_parts = fileBaseName.split("_")
    if name_parts[-1].isdigit():
        suffix_number = name_parts[-1]

    # get transition name
    name = fileBaseName.replace("_", " ").capitalize()

    # replace suffix number with placeholder (if any)
    if suffix_number:
        name = name.replace(suffix_number, "%s")

    # add text to list
    transitions_text[name] = full_subfile_path


log.info("-----------------------------------------------------")
log.info(" Creating the custom XML POT files")
log.info("-----------------------------------------------------")

# header of POT file
header_text = ""
header_text = header_text + '# OpenShot Video Editor POT Template File.\n'
header_text = header_text + '# Copyright (C) 2008-2018 OpenShot Studios, LLC\n'
header_text = header_text + '# This file is distributed under the same license as OpenShot.\n'
header_text = header_text + '# Jonathan Thomas <Jonathan.Oomph@gmail.com>, 2018.\n'
header_text = header_text + '#\n'
header_text = header_text + '#, fuzzy\n'
header_text = header_text + 'msgid ""\n'
header_text = header_text + 'msgstr ""\n'
header_text = header_text + '"Project-Id-Version: OpenShot Video Editor (version: %s)\\n"\n' % info.VERSION
header_text = header_text + '"Report-Msgid-Bugs-To: Jonathan Thomas <Jonathan.Oomph@gmail.com>\\n"\n'
header_text = header_text + '"POT-Creation-Date: %s\\n"\n' % datetime.datetime.now()
header_text = header_text + '"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\\n"\n'
header_text = header_text + '"Last-Translator: Jonathan Thomas <Jonathan.Oomph@gmail.com>\\n"\n'
header_text = header_text + '"Language-Team: https://translations.launchpad.net/+groups/launchpad-translators\\n"\n'
header_text = header_text + '"MIME-Version: 1.0\\n"\n'
header_text = header_text + '"Content-Type: text/plain; charset=UTF-8\\n"\n'
header_text = header_text + '"Content-Transfer-Encoding: 8bit\\n"\n'

# Create POT files for the custom text (from our XML files)
temp_files = [['OpenShot_effects.pot', effects_text],
              ['OpenShot_export.pot', export_text],
              ['OpenShot_transitions.pot', transitions_text],
              ['OpenShot_emojis.pot', emoji_text],
              ['OpenShot_blender.pot', blender_text]
              ]
for temp_file, text_dict in temp_files:
    f = open(temp_file, "w")

    # write header
    f.write(header_text)

    # loop through each line of text
    for k, v in text_dict.items():
        if k:
            f.write('\n')
            f.write('#: %s\n' % v)
            f.write('msgid "%s"\n' % k)
            f.write('msgstr ""\n')

    # close file
    f.close()

log.info("-----------------------------------------------------")
log.info(" Combine all temp POT files using msgcat command ")
log.info(" (this removes dupes) ")
log.info("-----------------------------------------------------")

temp_files = ['OpenShot_source.pot', 'OpenShot_glade.pot', 'OpenShot_effects.pot',
              'OpenShot_export.pot', 'OpenShot_QtUi.pot']
command = "msgcat"
for temp_file in temp_files:
    # append files
    command = command + " " + os.path.join(language_folder_path, temp_file)
command = command + " -o " + os.path.join(language_folder_path, "OpenShot", "OpenShot.pot")

log.info(command)

# merge all 4 temp POT files
subprocess.call(command, shell=True)

log.info("-----------------------------------------------------")
log.info(" Create FINAL POT File from all temp POT files ")
log.info("-----------------------------------------------------")

# get the entire text of OpenShot.POT
f = open(os.path.join(language_folder_path, "OpenShot", "OpenShot.pot"), "r")
# read entire text of file
entire_source = f.read()
f.close()

# Create Final POT Output File
if os.path.exists(os.path.join(language_folder_path, "OpenShot", "OpenShot.pot")):
    os.remove(os.path.join(language_folder_path, "OpenShot", "OpenShot.pot"))
final = open(os.path.join(language_folder_path, "OpenShot", "OpenShot.pot"), "w")
final.write(header_text)
final.write("\n")

# Move transitions POT file to final location
if os.path.exists(os.path.join(language_folder_path, "OpenShot_transitions.pot")):
    os.rename(os.path.join(language_folder_path, "OpenShot_transitions.pot"),
              os.path.join(language_folder_path, "OpenShot", "OpenShot_transitions.pot"))

# Move emoji POT file to final location
if os.path.exists(os.path.join(language_folder_path, "OpenShot_emojis.pot")):
    os.rename(os.path.join(language_folder_path, "OpenShot_emojis.pot"),
              os.path.join(language_folder_path, "OpenShot", "OpenShot_emojis.pot"))

# Move blender POT file to final location
if os.path.exists(os.path.join(language_folder_path, "OpenShot_blender.pot")):
    os.rename(os.path.join(language_folder_path, "OpenShot_blender.pot"),
              os.path.join(language_folder_path, "OpenShot", "OpenShot_blender.pot"))

# Trim the beginning off of each POT file
start_pos = entire_source.find("#: ")
trimmed_source = entire_source[start_pos:]

# Add to Final POT File
final.write(trimmed_source)
final.write("\n")

# Close final POT file
final.close()

log.info("-----------------------------------------------------")
log.info(" Remove all temp POT files ")
log.info("-----------------------------------------------------")

# Delete all 4 temp files
temp_files = ['OpenShot_source.pot', 'OpenShot_glade.pot', 'OpenShot_effects.pot', 'OpenShot_export.pot',
              'OpenShot_transitions.pot', 'OpenShot_QtUi.pot', 'OpenShot_QtUi.ts']
for temp_file_name in temp_files:
    temp_file_path = os.path.join(language_folder_path, temp_file_name)
    if os.path.exists(temp_file_path):
        os.remove(temp_file_path)

# output success
log.info("-----------------------------------------------------")
log.info(" The OpenShot.pot file has been successfully created ")
log.info(" with all text in OpenShot.")
log.info("")
log.info(" Checking for duplicate keys...")
log.info("-----------------------------------------------------")

# Find any duplicate translations between our 4 template files
# If these duplicates are translated differently, we will end up
# with conflicts, and both translations will be combined incorrectly
all_strings = {}

for pot_file in [
    'OpenShot.pot',
    'OpenShot_transitions.pot',
    'OpenShot_blender.pot',
    'OpenShot_emojis.pot',
]:
    with open(os.path.join(language_folder_path, "OpenShot", pot_file)) as f:
        data = f.read()
        for key in re.findall('^msgid \"(.*)\"', data, re.MULTILINE):
            if key not in all_strings:
                all_strings[key] = "%s | %s" % (key, pot_file)
            elif key and key not in ('translator-credits', ):
                log.info(' ERROR: Duplicate key found: %s::%s' % (pot_file, all_strings[key]))



"""
 @file
 @brief This file contains a lightweight in-memory generation queue for ComfyUI jobs.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import uuid
from collections import deque
from threading import Event
from time import monotonic

from PyQt5.QtCore import QObject, QThread, pyqtSignal, pyqtSlot

from classes.comfy_client import ComfyClient
from classes.logger import log


class _GenerationWorker(QObject):
    """Background worker that simulates generation progress for queued jobs."""

    progress_changed = pyqtSignal(str, int)
    progress_detail_changed = pyqtSignal(str, str)
    progress_sub_changed = pyqtSignal(str, int)
    job_finished = pyqtSignal(str, bool, bool, str, object)

    def __init__(self):
        super().__init__()
        self._cancel_requested = set()
        self._job_prompts = {}

    def _is_cancel_requested(self, job_id, cancel_event):
        return (job_id in self._cancel_requested) or (cancel_event is not None and cancel_event.is_set())

    @staticmethod
    def _is_unfinished_meta_batch(history_entry):
        outputs = history_entry.get("outputs", {}) if isinstance(history_entry, dict) else {}
        if not isinstance(outputs, dict):
            return False
        for node_out in outputs.values():
            if not isinstance(node_out, dict):
                continue
            unfinished = node_out.get("unfinished_batch", None)
            if isinstance(unfinished, list):
                if any(bool(v) for v in unfinished):
                    return True
            elif unfinished:
                return True
        return False

    @staticmethod
    def _history_prompt_meta(history_entry):
        prompt_payload = history_entry.get("prompt", []) if isinstance(history_entry, dict) else []
        if not isinstance(prompt_payload, list):
            return "", 0
        client_payload = prompt_payload[3] if len(prompt_payload) >= 4 else {}
        if not isinstance(client_payload, dict):
            return "", 0
        client_id = str(client_payload.get("client_id", "")).strip()
        create_time = int(client_payload.get("create_time", 0) or 0)
        return client_id, create_time

    @staticmethod
    def _allow_unfiltered_output_fallback(template_id):
        template_id = str(template_id or "").strip().lower()
        # Track-object templates intentionally have multiple save nodes
        # (mask/debug + final), so we must not relax save-node filtering.
        if template_id in (
            "video-blur-anything-sam2",
            "video-mask-anything-sam2",
            "video-highlight-anything-sam2",
            "txt2music-ace-step",
        ):
            return False
        return True

    def _find_related_meta_batch_outputs(self, client, history_entry, save_node_ids, template_id=""):
        base_client_id, base_create_time = self._history_prompt_meta(history_entry)
        if not base_client_id:
            return []
        try:
            history_all = client.history_all() or {}
        except Exception:
            return []
        if not isinstance(history_all, dict):
            return []

        best_create_time = 0
        best_outputs = []
        for entry in history_all.values():
            if not isinstance(entry, dict):
                continue
            status_obj = entry.get("status", {}) if isinstance(entry, dict) else {}
            status_str = str(status_obj.get("status_str", "")).lower()
            if status_str not in ("success", "completed", ""):
                continue
            entry_client_id, entry_create_time = self._history_prompt_meta(entry)
            if entry_client_id != base_client_id:
                continue
            if entry_create_time and base_create_time and entry_create_time < base_create_time:
                continue
            if self._is_unfinished_meta_batch(entry):
                continue

            outputs = ComfyClient.extract_file_outputs(entry, save_node_ids=save_node_ids)
            if (not outputs) and save_node_ids and self._allow_unfiltered_output_fallback(template_id):
                outputs = ComfyClient.extract_file_outputs(entry, save_node_ids=None)
            if not outputs:
                continue

            if entry_create_time >= best_create_time:
                best_create_time = entry_create_time
                best_outputs = outputs

        return best_outputs

    @pyqtSlot(str, object)
    def run_job(self, job_id, request):
        request = request or {}
        cancel_event = request.get("cancel_event")
        if request.get("workflow") and request.get("comfy_url"):
            self._run_comfy_job(job_id, request)
            return

        canceled = False
        for step in range(1, 21):
            QThread.msleep(250)
            if self._is_cancel_requested(job_id, cancel_event):
                canceled = True
                break
            self.progress_changed.emit(job_id, min(step * 5, 99))

        self._cancel_requested.discard(job_id)

        if canceled:
            self.job_finished.emit(job_id, False, True, "", [])
        else:
            self.progress_changed.emit(job_id, 100)
            self.job_finished.emit(job_id, True, False, "", [])

    @pyqtSlot(str)
    def cancel_job(self, job_id):
        log.debug("GenerationWorker cancel_job received job=%s", str(job_id))
        self._cancel_requested.add(job_id)

    def _run_comfy_job(self, job_id, request):
        comfy_url = request.get("comfy_url")
        workflow = request.get("workflow")
        client_id = request.get("client_id") or "openshot-qt"
        timeout_s = int(request.get("timeout_s") or 86400)  # default 24 hours safety cap
        save_node_ids = list(request.get("save_node_ids") or [])
        template_id = str(request.get("template_id") or "")
        cancel_event = request.get("cancel_event")
        client = ComfyClient(comfy_url)
        ws_client = None

        try:
            prompt_id = client.queue_prompt(workflow, client_id)
            if not prompt_id:
                self.job_finished.emit(job_id, False, False, "ComfyUI returned an invalid prompt_id", [])
                return
            self._job_prompts[job_id] = prompt_id
            try:
                ws_client = ComfyClient.open_progress_socket(comfy_url, client_id)
                log.debug("Comfy progress websocket connected for prompt=%s", str(prompt_id))
            except Exception:
                log.debug("Comfy progress websocket unavailable; continuing without live progress", exc_info=True)

            start_time = monotonic()
            last_in_queue_time = start_time
            last_contact_time = start_time
            last_progress_log_time = 0.0
            last_network_error_log_time = 0.0
            progress_endpoint_unavailable = False
            accepted_progress_started = False
            ws_retry_delay_s = 2.0
            ws_next_retry_at = start_time
            ws_last_progress_time = start_time
            ws_stale_reconnect_s = 60.0
            ws_stale_reconnect_max_s = 300.0
            prompt_key = str(prompt_id)
            last_progress_signature = None
            last_progress_detail = ""

            while True:
                if self._is_cancel_requested(job_id, cancel_event):
                    log.debug("Comfy cancel requested for job=%s prompt=%s", job_id, str(prompt_id))
                    cancel_ok = False
                    cancel_errors = []

                    # Retry cancellation a few times and verify prompt no longer appears in Comfy queue.
                    for attempt in range(1, 181):
                        try:
                            cancel_ok = client.cancel_prompt(prompt_id) or cancel_ok
                        except Exception as ex:
                            cancel_errors.append("queue: {}".format(ex))

                        try:
                            cancel_ok = client.interrupt(prompt_id=prompt_id) or cancel_ok
                        except Exception as ex:
                            cancel_errors.append("interrupt: {}".format(ex))

                        try:
                            history = client.history(prompt_id) or {}
                            prompt_key = str(prompt_id)
                            history_entry = history.get(prompt_key) or history.get(prompt_id) or None
                            if isinstance(history_entry, dict):
                                status_obj = history_entry.get("status", {}) if isinstance(history_entry, dict) else {}
                                status_str = str(status_obj.get("status_str", "")).lower()
                                # Comfy commonly marks interrupted runs as failed/error in history.
                                if status_str in ("error", "failed"):
                                    cancel_ok = True
                                    log.debug(
                                        "Comfy cancel confirmed by history status for job=%s prompt=%s status=%s",
                                        job_id,
                                        prompt_key,
                                        status_str,
                                    )
                                    break
                        except Exception as ex:
                            cancel_errors.append("history-check: {}".format(ex))

                        try:
                            queue_data = client.queue() or {}
                            if not ComfyClient.prompt_in_queue(prompt_id, queue_data):
                                cancel_ok = True
                                log.debug(
                                    "Comfy cancel confirmed by queue absence for job=%s prompt=%s on attempt=%s",
                                    job_id,
                                    str(prompt_id),
                                    attempt,
                                )
                                break
                        except Exception as ex:
                            cancel_errors.append("queue-check: {}".format(ex))

                        if attempt % 10 == 0:
                            log.debug(
                                "Comfy cancel still pending job=%s prompt=%s attempt=%s",
                                job_id,
                                str(prompt_id),
                                attempt,
                            )
                        QThread.msleep(500)

                    self._cancel_requested.discard(job_id)
                    self._job_prompts.pop(job_id, None)
                    if cancel_ok:
                        self.job_finished.emit(job_id, False, True, "", [])
                    else:
                        self.job_finished.emit(
                            job_id,
                            False,
                            False,
                            "ComfyUI did not accept cancel request ({})".format("; ".join(cancel_errors) or "unknown"),
                            [],
                        )
                    return

                history_entry = None
                try:
                    history = client.history(prompt_id) or {}
                    history_entry = history.get(prompt_key) or history.get(prompt_id) or None
                    last_contact_time = monotonic()
                except Exception:
                    now_log = monotonic()
                    if (now_log - last_network_error_log_time) > 8.0:
                        log.debug(
                            "Comfy history poll temporarily unavailable for job=%s prompt=%s; retrying",
                            job_id,
                            prompt_key,
                            exc_info=True,
                        )
                        last_network_error_log_time = now_log
                if history_entry is not None:
                    status_obj = history_entry.get("status", {}) if isinstance(history_entry, dict) else {}
                    status_str = str(status_obj.get("status_str", "")).lower()
                    if status_str in ("error", "failed"):
                        error_text = "ComfyUI job failed."
                        messages = status_obj.get("messages", [])
                        if isinstance(messages, list) and messages:
                            error_text = ComfyClient.summarize_error_text(messages[-1])
                        self._job_prompts.pop(job_id, None)
                        self.job_finished.emit(job_id, False, False, error_text, [])
                        return
                    if self._is_unfinished_meta_batch(history_entry):
                        image_outputs = self._find_related_meta_batch_outputs(
                            client,
                            history_entry,
                            save_node_ids,
                            template_id=template_id,
                        )
                        if image_outputs:
                            self.progress_changed.emit(job_id, 100)
                            self._job_prompts.pop(job_id, None)
                            self.job_finished.emit(job_id, True, False, "", image_outputs)
                            return
                        # Meta batch uses follow-up prompts under the same client_id.
                        # Keep polling progress/queue while waiting for follow-up prompt outputs.
                    else:
                        image_outputs = ComfyClient.extract_file_outputs(history_entry, save_node_ids=save_node_ids)
                        if (not image_outputs) and save_node_ids and self._allow_unfiltered_output_fallback(template_id):
                            # Fallback for workflows whose output node ids shift or emit non-standard keys.
                            image_outputs = ComfyClient.extract_file_outputs(history_entry, save_node_ids=None)
                        self.progress_changed.emit(job_id, 100)
                        self._job_prompts.pop(job_id, None)
                        self.job_finished.emit(job_id, True, False, "", image_outputs)
                        return

                # Query ComfyUI's live progress values when available.
                try:
                    ws_progress_emitted = False
                    now = monotonic()
                    if ws_client is None and now >= ws_next_retry_at:
                        try:
                            ws_client = ComfyClient.open_progress_socket(comfy_url, client_id)
                            ws_retry_delay_s = 2.0
                            log.debug("Comfy progress websocket reconnected for prompt=%s", prompt_key)
                        except Exception:
                            ws_next_retry_at = now + ws_retry_delay_s
                            ws_retry_delay_s = min(60.0, ws_retry_delay_s * 1.5)
                            now_log = monotonic()
                            if (now_log - last_network_error_log_time) > 8.0:
                                log.debug(
                                    "Comfy websocket reconnect failed for job=%s prompt=%s; retrying in %.1fs",
                                    job_id,
                                    prompt_key,
                                    ws_retry_delay_s,
                                    exc_info=True,
                                )
                                last_network_error_log_time = now_log

                    if ws_client is not None:
                        try:
                            # Accept progress from follow-up prompts as well (meta-batch).
                            progress_event = ws_client.poll_progress(prompt_id=None)
                        except Exception:
                            progress_event = None
                            try:
                                ws_client.close()
                            except Exception:
                                pass
                            ws_client = None
                            ws_next_retry_at = monotonic() + ws_retry_delay_s
                            ws_retry_delay_s = min(60.0, ws_retry_delay_s * 1.5)
                            now_log = monotonic()
                            if (now_log - last_network_error_log_time) > 8.0:
                                log.debug(
                                    "Comfy websocket progress read failed for job=%s prompt=%s; switching to retry mode",
                                    job_id,
                                    prompt_key,
                                    exc_info=True,
                                )
                                last_network_error_log_time = now_log

                        if progress_event is not None:
                            elapsed = monotonic() - start_time
                            progress = int(progress_event.get("percent", 0))
                            raw_value = float(progress_event.get("value", 0.0))
                            raw_max = float(progress_event.get("max", 0.0))
                            progress_type = str(progress_event.get("type", ""))
                            progress_node = str(progress_event.get("node", ""))
                            # Some workflows emit near-complete progress bursts at startup
                            # (e.g. tiny setup nodes), then reset to sampler progress.
                            # Ignore those bootstrap spikes for a short window.
                            if (
                                (not accepted_progress_started)
                                and progress >= 95
                                and elapsed < 20.0
                                and raw_max <= 1.0
                            ):
                                log.debug(
                                    "Comfy WS progress setup-node spike ignored job=%s prompt=%s node=%s type=%s value=%s max=%s percent=%s elapsed=%.2fs",
                                    job_id,
                                    prompt_key,
                                    progress_node,
                                    progress_type,
                                    raw_value,
                                    raw_max,
                                    progress,
                                    elapsed,
                                )
                            elif (not accepted_progress_started) and progress >= 95 and elapsed < 20.0:
                                log.debug(
                                    "Comfy WS progress bootstrap spike ignored job=%s prompt=%s percent=%s elapsed=%.2fs",
                                    job_id,
                                    prompt_key,
                                    progress,
                                    elapsed,
                                )
                            else:
                                accepted_progress_started = True
                                progress_signature = (
                                    progress_type,
                                    progress_node,
                                    int(progress),
                                    round(raw_value, 3),
                                    round(raw_max, 3),
                                )
                                if progress_signature != last_progress_signature:
                                    inferred_progress = int(max(0, min(99, progress)))
                                    detail_text = ""
                                    if progress_node:
                                        detail_text = "node {} {}%".format(progress_node, int(progress))

                                    log.debug(
                                        "Comfy WS progress emit job=%s prompt=%s node=%s type=%s value=%s max=%s percent=%s",
                                        job_id,
                                        prompt_key,
                                        progress_node,
                                        progress_type,
                                        raw_value,
                                        raw_max,
                                        inferred_progress,
                                    )
                                    self.progress_changed.emit(job_id, inferred_progress)
                                    self.progress_sub_changed.emit(job_id, int(max(0, min(99, progress))))
                                    if detail_text != last_progress_detail:
                                        self.progress_detail_changed.emit(job_id, detail_text)
                                        last_progress_detail = detail_text
                                    last_progress_signature = progress_signature
                                ws_progress_emitted = True
                                ws_last_progress_time = monotonic()
                                ws_stale_reconnect_s = 60.0
                                last_contact_time = monotonic()
                    if ws_client is not None and not ws_progress_emitted:
                        stale_for = now - ws_last_progress_time
                        if stale_for >= ws_stale_reconnect_s:
                            try:
                                ws_client.close()
                            except Exception:
                                pass
                            ws_client = None
                            ws_next_retry_at = now + ws_retry_delay_s
                            ws_retry_delay_s = min(60.0, ws_retry_delay_s * 1.5)
                            next_stale_reconnect_s = min(
                                ws_stale_reconnect_max_s,
                                max(60.0, ws_stale_reconnect_s * 1.5),
                            )
                            log.debug(
                                "Comfy websocket stalled for job=%s prompt=%s (%.1fs >= %.1fs); forcing reconnect, next stall timeout %.1fs",
                                job_id,
                                prompt_key,
                                stale_for,
                                ws_stale_reconnect_s,
                                next_stale_reconnect_s,
                            )
                            ws_stale_reconnect_s = next_stale_reconnect_s
                    # Use HTTP /progress only when websocket progress is unavailable.
                    # If websocket is connected but temporarily quiet, keep waiting for WS
                    # instead of spamming a misleading 404 fallback warning.
                    if ws_client is None:
                        progress_data = client.progress()
                        if progress_data is None:
                            if not progress_endpoint_unavailable:
                                log.debug(
                                    "Comfy progress endpoint unavailable (404); waiting for websocket progress for job=%s",
                                    job_id,
                                )
                            progress_endpoint_unavailable = True
                            progress_data = {}

                        progress_block = progress_data.get("progress", progress_data)
                        if not isinstance(progress_block, dict):
                            progress_block = {}

                        value = float(progress_block.get("value", progress_block.get("current", 0.0)))
                        maximum = float(progress_block.get("max", progress_block.get("total", 0.0)))
                        progress_prompt = str(
                            progress_data.get("prompt_id", progress_block.get("prompt_id", ""))
                        )
                        prompt_matches = (not progress_prompt) or (progress_prompt == prompt_key)

                        now_log = monotonic()
                        if (now_log - last_progress_log_time) > 8.0:
                            log.debug(
                                "Comfy progress poll job=%s prompt=%s payload_keys=%s value=%s max=%s progress_prompt=%s prompt_match=%s",
                                job_id,
                                prompt_key,
                                list(progress_data.keys()) if isinstance(progress_data, dict) else type(progress_data),
                                value,
                                maximum,
                                progress_prompt,
                                prompt_matches,
                            )
                            last_progress_log_time = now_log

                        if maximum > 0 and prompt_matches:
                            progress = int(max(0, min(99, round((value / maximum) * 100.0))))
                            progress_signature = ("poll", "", int(progress), round(value, 3), round(maximum, 3))
                            if progress_signature != last_progress_signature:
                                log.debug(
                                    "Comfy progress emit job=%s prompt=%s value=%s max=%s percent=%s",
                                    job_id,
                                    prompt_key,
                                    value,
                                    maximum,
                                    progress,
                                )
                                self.progress_changed.emit(job_id, progress)
                                self.progress_sub_changed.emit(job_id, int(max(0, min(99, progress))))
                                if last_progress_detail:
                                    self.progress_detail_changed.emit(job_id, "")
                                    last_progress_detail = ""
                                last_progress_signature = progress_signature
                            last_contact_time = monotonic()
                except Exception:
                    # Keep polling history and queue even if /progress is unavailable.
                    now_log = monotonic()
                    if (now_log - last_network_error_log_time) > 8.0:
                        log.debug("Comfy progress poll failed for job=%s", job_id, exc_info=True)
                        last_network_error_log_time = now_log

                # Check queue to avoid timing out long-running but active jobs.
                in_queue = False
                try:
                    queue_data = client.queue() or {}
                    in_queue = ComfyClient.prompt_in_queue(prompt_id, queue_data)
                    last_contact_time = monotonic()
                except Exception:
                    # If queue check fails, do not penalize the job immediately.
                    in_queue = True
                    now_log = monotonic()
                    if (now_log - last_network_error_log_time) > 8.0:
                        log.debug("Comfy queue check temporarily unavailable for job=%s", job_id, exc_info=True)
                        last_network_error_log_time = now_log
                if in_queue:
                    last_in_queue_time = monotonic()
                else:
                    now_log = monotonic()
                    if (now_log - last_progress_log_time) > 8.0:
                        log.debug(
                            "Comfy queue check: prompt=%s not found in queue_running/queue_pending yet",
                            prompt_key,
                        )
                        last_progress_log_time = now_log

                now = monotonic()
                if (now - start_time) > timeout_s:
                    self._job_prompts.pop(job_id, None)
                    self.job_finished.emit(job_id, False, False, "Timed out waiting for ComfyUI history result", [])
                    return

                if (now - last_contact_time) > 60.0:
                    now_log = monotonic()
                    if (now_log - last_network_error_log_time) > 8.0:
                        log.debug(
                            "Comfy connection degraded for job=%s prompt=%s (no successful API contact for %.1fs); continuing retries",
                            job_id,
                            prompt_key,
                            now - last_contact_time,
                        )
                        last_network_error_log_time = now_log

                # If prompt vanished from queue for an extended period and still no history, treat as failure.
                if (now - last_in_queue_time) > 600:
                    self._job_prompts.pop(job_id, None)
                    self.job_finished.emit(
                        job_id,
                        False,
                        False,
                        "ComfyUI prompt is no longer in queue and has no history result.",
                        [],
                    )
                    return
                QThread.msleep(500)
        except Exception as ex:
            self._job_prompts.pop(job_id, None)
            self.job_finished.emit(job_id, False, False, ComfyClient.summarize_error_text(ex), [])
        finally:
            if ws_client is not None:
                ws_client.close()


class GenerationQueueManager(QObject):
    """Single-worker, in-memory generation queue with per-file active-job limits."""

    ACTIVE_STATES = {"queued", "running", "canceling"}

    job_added = pyqtSignal(str, object)
    job_updated = pyqtSignal(str, str, int)
    job_finished = pyqtSignal(str, str)
    job_removed = pyqtSignal(str)
    file_job_changed = pyqtSignal(str)
    queue_changed = pyqtSignal()

    _run_job = pyqtSignal(str, object)
    _cancel_job = pyqtSignal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.jobs = {}
        self._queued = deque()
        self._running_job_id = None
        self._active_file_jobs = {}

        self._thread = QThread(self)
        self._thread.setObjectName("generation_queue_worker")
        self._worker = _GenerationWorker()
        self._worker.moveToThread(self._thread)
        self._run_job.connect(self._worker.run_job)
        self._cancel_job.connect(self._worker.cancel_job)
        self._worker.progress_changed.connect(self._on_progress_changed)
        self._worker.progress_detail_changed.connect(self._on_progress_detail_changed)
        self._worker.progress_sub_changed.connect(self._on_progress_sub_changed)
        self._worker.job_finished.connect(self._on_job_finished)
        self._thread.start()

    def enqueue(self, name, template_id, prompt, source_file_id=None, request=None):
        source_file_id = str(source_file_id or "")
        if source_file_id and self.get_active_job_for_file(source_file_id):
            return None

        job_id = str(uuid.uuid4())
        cancel_event = Event()
        job_request = dict(request or {})
        job_request["cancel_event"] = cancel_event
        job = {
            "id": job_id,
            "name": str(name or "").strip(),
            "template_id": str(template_id or "").strip(),
            "prompt": str(prompt or "").strip(),
            "source_file_id": source_file_id,
            "status": "queued",
            "progress": 0,
            "sub_progress": 0,
            "progress_detail": "",
            "error": "",
            "request": job_request,
            "cancel_event": cancel_event,
        }
        self.jobs[job_id] = job
        self._queued.append(job_id)
        if source_file_id:
            self._active_file_jobs[source_file_id] = job_id

        self.job_added.emit(job_id, source_file_id)
        self.job_updated.emit(job_id, "queued", 0)
        self._emit_file_changed(source_file_id)
        self.queue_changed.emit()
        self._start_next_if_idle()
        return job_id

    def cancel_job(self, job_id):
        job = self.jobs.get(job_id)
        if not job:
            log.debug("GenerationQueue cancel_job ignored; unknown job=%s", str(job_id))
            return False

        log.debug(
            "GenerationQueue cancel_job request job=%s status=%s source_file_id=%s",
            str(job_id),
            str(job.get("status", "")),
            str(job.get("source_file_id", "")),
        )
        if job["status"] == "queued":
            cancel_event = job.get("cancel_event")
            if cancel_event is not None:
                cancel_event.set()
                log.debug("GenerationQueue cancel_event set for queued job=%s", str(job_id))
            job["status"] = "canceled"
            self._queued = deque([queued_id for queued_id in self._queued if queued_id != job_id])
            self._release_file_slot(job.get("source_file_id", ""))
            self.job_updated.emit(job_id, "canceled", int(job.get("progress", 0)))
            self.job_finished.emit(job_id, "canceled")
            self._emit_file_changed(job.get("source_file_id", ""))
            self.queue_changed.emit()
            log.debug("GenerationQueue cancel_job completed for queued job=%s", str(job_id))
            return True

        if job["status"] == "running":
            cancel_event = job.get("cancel_event")
            if cancel_event is not None:
                cancel_event.set()
                log.debug("GenerationQueue cancel_event set for running job=%s", str(job_id))
            job["status"] = "canceling"
            self.job_updated.emit(job_id, "canceling", int(job.get("progress", 0)))
            self._cancel_job.emit(job_id)
            self._emit_file_changed(job.get("source_file_id", ""))
            self.queue_changed.emit()
            log.debug("GenerationQueue cancel_job emitted worker cancel for running job=%s", str(job_id))
            return True

        log.debug("GenerationQueue cancel_job ignored for job=%s with status=%s", str(job_id), str(job.get("status", "")))
        return False

    def cancel_jobs_for_file(self, source_file_id):
        source_file_id = str(source_file_id or "")
        if not source_file_id:
            return
        for job in list(self.jobs.values()):
            if job.get("source_file_id") == source_file_id and job.get("status") in self.ACTIVE_STATES:
                self.cancel_job(job["id"])

    def remove_job(self, job_id):
        job = self.jobs.get(job_id)
        if not job:
            return False
        if job.get("status") in self.ACTIVE_STATES:
            return False

        source_file_id = job.get("source_file_id", "")
        self.jobs.pop(job_id, None)
        self.job_removed.emit(job_id)
        self._emit_file_changed(source_file_id)
        self.queue_changed.emit()
        return True

    def get_job(self, job_id):
        return self.jobs.get(job_id)

    def get_active_job_for_file(self, source_file_id):
        source_file_id = str(source_file_id or "")
        if not source_file_id:
            return None

        job_id = self._active_file_jobs.get(source_file_id)
        if not job_id:
            return None

        job = self.jobs.get(job_id)
        if not job or job.get("status") not in self.ACTIVE_STATES:
            self._active_file_jobs.pop(source_file_id, None)
            return None
        return job

    def get_file_badge(self, source_file_id):
        job = self.get_active_job_for_file(source_file_id)
        if not job:
            return None

        status = job.get("status")
        progress = int(job.get("progress", 0))
        sub_progress = int(job.get("sub_progress", 0))
        detail = str(job.get("progress_detail", "") or "").strip()
        if status == "queued":
            label = "Queued"
        elif status == "running":
            label = "Generating {}%".format(progress)
            if detail:
                label = "{} ({})".format(label, detail)
        elif status == "canceling":
            label = "Canceling..."
        else:
            label = status.capitalize()

        return {
            "status": status,
            "progress": progress,
            "sub_progress": sub_progress,
            "label": label,
            "job_id": job.get("id"),
        }

    def shutdown(self):
        if self._thread.isRunning():
            self._thread.quit()
            self._thread.wait(2000)

    def _start_next_if_idle(self):
        if self._running_job_id is not None:
            return
        if not self._queued:
            return

        next_job_id = self._queued.popleft()
        job = self.jobs.get(next_job_id)
        if not job:
            self._start_next_if_idle()
            return

        self._running_job_id = next_job_id
        job["status"] = "running"
        job["progress"] = int(job.get("progress", 0))
        job["sub_progress"] = int(job.get("sub_progress", 0))
        self.job_updated.emit(next_job_id, "running", int(job["progress"]))
        self._emit_file_changed(job.get("source_file_id", ""))
        self.queue_changed.emit()
        self._run_job.emit(next_job_id, job.get("request", {}))

    def _release_file_slot(self, source_file_id):
        source_file_id = str(source_file_id or "")
        if source_file_id:
            self._active_file_jobs.pop(source_file_id, None)

    def _emit_file_changed(self, source_file_id):
        source_file_id = str(source_file_id or "")
        if source_file_id:
            self.file_job_changed.emit(source_file_id)

    @pyqtSlot(str, int)
    def _on_progress_changed(self, job_id, progress):
        job = self.jobs.get(job_id)
        if not job:
            return
        if job.get("status") not in ("running", "canceling"):
            return
        job["progress"] = int(progress)
        self.job_updated.emit(job_id, job.get("status"), int(progress))
        self._emit_file_changed(job.get("source_file_id", ""))
        self.queue_changed.emit()

    @pyqtSlot(str, str)
    def _on_progress_detail_changed(self, job_id, detail):
        job = self.jobs.get(job_id)
        if not job:
            return
        if job.get("status") not in ("running", "canceling"):
            return
        detail_text = str(detail or "").strip()
        if str(job.get("progress_detail", "") or "") == detail_text:
            return
        job["progress_detail"] = detail_text
        self.job_updated.emit(job_id, job.get("status"), int(job.get("progress", 0)))
        self._emit_file_changed(job.get("source_file_id", ""))
        self.queue_changed.emit()

    @pyqtSlot(str, int)
    def _on_progress_sub_changed(self, job_id, progress):
        job = self.jobs.get(job_id)
        if not job:
            return
        if job.get("status") not in ("running", "canceling"):
            return
        p = int(max(0, min(99, progress)))
        if int(job.get("sub_progress", 0)) == p:
            return
        job["sub_progress"] = p
        self.job_updated.emit(job_id, job.get("status"), int(job.get("progress", 0)))
        self._emit_file_changed(job.get("source_file_id", ""))
        self.queue_changed.emit()

    @pyqtSlot(str, bool, bool, str, object)
    def _on_job_finished(self, job_id, success, canceled, error, outputs):
        job = self.jobs.get(job_id)
        if not job:
            return

        if canceled:
            job["status"] = "canceled"
        elif success:
            job["status"] = "completed"
            job["progress"] = 100
            job["outputs"] = list(outputs or [])
        else:
            job["status"] = "failed"
            job["error"] = str(error or "")

        source_file_id = job.get("source_file_id", "")
        self._release_file_slot(source_file_id)

        self.job_updated.emit(job_id, job["status"], int(job.get("progress", 0)))
        self.job_finished.emit(job_id, job["status"])
        self._emit_file_changed(source_file_id)
        self.queue_changed.emit()

        if self._running_job_id == job_id:
            self._running_job_id = None
        self._start_next_if_idle()



"""
 @file
 @brief This file listens to changes, and updates the primary project data
 @author Noah Figg <eggmunkee@hotmail.com>
 @author Jonathan Thomas <jonathan@openshot.org>
 @author Olivier Girard <eolinwen@gmail.com>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import copy
import glob
import os
import random
import shutil
import json

from classes import info
from classes.app import get_app
from classes.image_types import get_media_type
from classes.json_data import JsonDataStore
from classes.logger import log
from classes.thumbnail import MigrateThumbnailLayout
from classes.updates import UpdateInterface
from classes.assets import get_assets_path
from classes.path_utils import comparable_local_path, normalized_local_path
from windows.views.find_file import find_missing_file
from classes.convert_framerate import change_profile

from .keyframe_scaler import KeyframeScaler

import openshot


class ProjectDataStore(JsonDataStore, UpdateInterface):
    """ This class allows advanced searching of data structure, implements changes interface """

    def __init__(self):
        JsonDataStore.__init__(self)
        self.data_type = "project data"  # Used in error messages
        self.default_project_filepath = os.path.join(info.PATH, 'settings', '_default.project')

        # Set default filepath to user's home folder
        self.current_filepath = None

        # Track changes after save
        self.has_unsaved_changes = False

        # Load default project data on creation
        self.new()

    def needs_save(self):
        """Returns if project data has unsaved changes"""
        return self.has_unsaved_changes

    def _effect_has_reader_source(self, effect):
        """Return True when an effect already has a modern reader payload."""
        if not isinstance(effect, dict):
            return False
        for key in ("mask_reader", "reader"):
            reader = effect.get(key)
            if not isinstance(reader, dict):
                continue
            if reader.get("path") or reader.get("id") or reader.get("has_single_image"):
                return True
        return False

    def _migrate_optimized_asset_paths(self):
        """Migrate legacy proxy asset folder paths from `proxies` to `optimized`."""
        if not self.current_filepath:
            return

        asset_path = get_assets_path(self.current_filepath)
        legacy_proxy_path = os.path.join(asset_path, "proxies")
        optimized_path = os.path.join(asset_path, "optimized")

        if os.path.isdir(legacy_proxy_path) and not os.path.exists(optimized_path):
            shutil.move(legacy_proxy_path, optimized_path)
            log.info("Migrated optimized assets folder to %s", optimized_path)

        legacy_prefix = os.path.abspath(legacy_proxy_path) + os.sep
        updated = 0
        for file_data in self._data.get("files", []):
            if not isinstance(file_data, dict):
                continue
            proxy_reader = file_data.get("proxy_reader")
            if not isinstance(proxy_reader, dict):
                continue
            proxy_path = str(proxy_reader.get("path") or "")
            abs_proxy_path = os.path.abspath(proxy_path) if proxy_path else ""
            if abs_proxy_path.startswith(legacy_prefix):
                relative_path = os.path.relpath(abs_proxy_path, os.path.abspath(legacy_proxy_path))
                proxy_reader["path"] = os.path.join(optimized_path, relative_path)
                updated += 1

        info.PROXY_PATH = optimized_path
        if updated:
            log.info("Updated %s optimized reader path(s) to use %s", updated, optimized_path)

    def _drop_obsolete_effect_resource(self, effect):
        """Remove legacy resource paths once a reader payload is available."""
        if not isinstance(effect, dict):
            return
        if "resource" in effect and self._effect_has_reader_source(effect):
            effect.pop("resource", None)

    def get(self, key):
        """Get copied value of a given key in data store"""

        # Verify key is valid type
        if not key:
            log.warning("ProjectDataStore cannot get empty key.")
            return None
        if not isinstance(key, list):
            key = [key]

        # Get reference to internal data structure
        obj = self._data

        # Iterate through key list finding sub-objects either by name or by an object match criteria such as {"id":"ADB34"}.
        for key_index in range(len(key)):
            key_part = key[key_index]

            # Key_part must be a string or dictionary
            if not isinstance(key_part, dict) and not isinstance(key_part, str):
                log.error("Unexpected key part type: {}".format(type(key_part).__name__))
                return None

            # If key_part is a dictionary and obj is a list or dict, each key is tested as a property of the items in the current object
            # in the project data structure, and the first match is returned.
            if isinstance(key_part, dict) and isinstance(obj, list):
                # Overall status of finding a matching sub-object
                found = False
                # Loop through each item in object to find match
                for item_index in range(len(obj)):
                    item = obj[item_index]
                    # True until something disqualifies this as a match
                    match = True
                    # Check each key in key_part dictionary and if not found to be equal as a property in item, move on to next item in list
                    for subkey in key_part:
                        # Get each key in dictionary (i.e. "id", "layer", etc...)
                        subkey = subkey.lower()
                        # If object is missing the key or the values differ, then it doesn't match.
                        if not (subkey in item and item[subkey] == key_part[subkey]):
                            match = False
                            break
                    # If matched, set key_part to index of list or dict and stop loop
                    if match:
                        found = True
                        obj = item
                        break
                # No match found, return None
                if not found:
                    return None

            # If key_part is a string, homogenize to lower case for comparisons
            if isinstance(key_part, str):
                key_part = key_part.lower()

                # Check current obj type (should be dictionary)
                if not isinstance(obj, dict):
                    log.warn(
                        "Invalid project data structure. Trying to use a key on a non-dictionary object. Key part: {} (\"{}\").\nKey: {}".format(
                            (key_index), key_part, key))
                    return None

                # If next part of path isn't in current dictionary, return failure
                if key_part not in obj:
                    log.warn(
                        'Key not found in project. Mismatch on key part %s ("%s").\nKey: %s',
                        key_index, key_part, key)
                    return None

                # Get the matching item
                obj = obj[key_part]

        # After processing each key, we've found object, return it
        return obj

    def set(self, key, value):
        """Prevent calling JsonDataStore set() method. It is not allowed in ProjectDataStore, as changes come from UpdateManager."""
        raise RuntimeError("ProjectDataStore.set() is not allowed. Changes must route through UpdateManager.")

    def _set(self, key, values=None, add=False, remove=False):
        """ Store setting, but adding isn't allowed. All possible settings must be in default settings file. """

        log.debug(
            "_set key: %s, values: %s, add: %s, remove: %s",
            key, values, add, remove)
        parent, my_key = None, ""

        # Verify key is valid type
        if not isinstance(key, list):
            log.warning("_set() key must be a list. key=%s", key)
            return None
        if not key:
            log.warning("Cannot set empty key (key=%s)", key)
            return None

        # Get reference to internal data structure
        obj = self._data

        # Iterate through key list finding sub-objects either by name or by an object match criteria such as {"id":"ADB34"}.
        for key_index in range(len(key)):
            key_part = key[key_index]

            # Key_part must be a string or dictionary
            if not isinstance(key_part, dict) and not isinstance(key_part, str):
                log.error("Unexpected key part type: %s", type(key_part).__name__)
                return None

            # If key_part is a dictionary and obj is a list or dict, each key is tested as a property of the items in the current object
            # in the project data structure, and the first match is returned.
            if isinstance(key_part, dict) and isinstance(obj, list):
                # Overall status of finding a matching sub-object
                found = False
                # Loop through each item in object to find match
                for item_index in range(len(obj)):
                    item = obj[item_index]
                    # True until something disqualifies this as a match
                    match = True
                    # Check each key in key_part dictionary and if not found to be equal as a property in item, move on to next item in list
                    for subkey in key_part.keys():
                        # Get each key in dictionary (i.e. "id", "layer", etc...)
                        subkey = subkey.lower()
                        # If object is missing the key or the values differ, then it doesn't match.
                        if not (subkey in item and item[subkey] == key_part[subkey]):
                            match = False
                            break
                    # If matched, set key_part to index of list or dict and stop loop
                    if match:
                        found = True
                        obj = item
                        my_key = item_index
                        break
                # No match found, return None
                if not found:
                    return None


            # If key_part is a string, homogenize to lower case for comparisons
            if isinstance(key_part, str):
                key_part = key_part.lower()

                # Check current obj type (should be dictionary)
                if not isinstance(obj, dict):
                    return None

                # If next part of path isn't in current dictionary, return failure
                if key_part not in obj:
                    log.warn(
                        'Key not found in project. Mismatch on key part %s ("%s").\nKey: %s',
                        key_index, key_part, key)
                    return None

                # Get sub-object based on part key as new object, continue to next part
                obj = obj[key_part]
                my_key = key_part


            # Set parent to the last set obj (if not final iteration)
            if key_index < (len(key) - 1) or key_index == 0:
                parent = obj


        # After processing each key, we've found object and parent, return former value/s on update
        ret = json.loads(json.dumps(obj))

        # Apply the correct action to the found item
        if remove:
            del parent[my_key]

        else:

            # Add or Full Update
            # For adds to list perform an insert to index or the end if not specified
            if add and isinstance(parent, list):
                parent.append(values)

            # Otherwise, set the given index
            elif isinstance(values, dict):
                # Update existing dictionary value
                obj.update(values)

            else:

                # Update root string
                self._data[my_key] = values

        # Return the previous value to the matching item (used for history tracking)
        return ret

    # Load default project data
    def new(self):
        """ Try to load default project settings file, will raise error on failure """
        import openshot

        # Try to load user default project
        if os.path.exists(info.USER_DEFAULT_PROJECT):
            try:
                self._data = self.read_from_file(info.USER_DEFAULT_PROJECT)
            except (FileNotFoundError, PermissionError):
                log.warning(
                    "Unable to load user project defaults from %s",
                    info.USER_DEFAULT_PROJECT, exc_info=1)
            except Exception:
                raise
            else:
                log.info("Loaded user project defaults from %s",
                         info.USER_DEFAULT_PROJECT)
        else:
            # Fall back to OpenShot defaults, if user defaults didn't load
            self._data = self.read_from_file(self.default_project_filepath)

        self.current_filepath = None
        self.has_unsaved_changes = False

        # Reset info paths back to their default/initial values
        info.reset_userdirs()

        # Get default profile
        s = get_app().get_settings()
        default_profile_desc = s.get("default-profile")

        # Get profile (if any)
        profile = self.get_profile(profile_desc=default_profile_desc)
        if not profile:
            # Fallback, in case of missing/invalid default profile
            profile = self.get_profile(profile_desc="HD 720p 30 fps")

        # Update default profile (if it changed)
        if profile and default_profile_desc != profile.info.description:
            log.info(f"Updating default-profile from legacy `{default_profile_desc}` to `{profile.info.description}`.")
            s.set("default-profile", profile.info.description)

        # Apply default audio playback settings to this data structure
        self.apply_default_audio_settings()

        # Set default project ID
        self._data["id"] = self.generate_id()

    def get_profile(self, profile_desc=None, profile_key=None):
        """Attempt to find a specific profile"""
        profile = None

        # Loop through legacy profiles
        LEGACY_PROFILE_PATH = os.path.join(info.PROFILES_PATH, "legacy")
        legacy_profile = None
        legacy_profile_match = None  # Find the equivalent match of this legacy (if any)
        for legacy_filename in os.listdir(LEGACY_PROFILE_PATH):
            legacy_profile_path = os.path.join(LEGACY_PROFILE_PATH, legacy_filename)
            try:
                # Load Profile and append description
                temp_profile = openshot.Profile(legacy_profile_path)
                if profile_desc == temp_profile.info.description:
                    legacy_profile = temp_profile
                    break
            except RuntimeError:
                # Ignore legacy parsing errors
                pass

        # Loop through profiles
        profile_dirs = [info.USER_PROFILES_PATH, info.PROFILES_PATH]
        available_dirs = [f for f in profile_dirs if os.path.exists(f)]
        for profile_folder in available_dirs:
            for file in reversed(sorted(os.listdir(profile_folder))):
                profile_path = os.path.join(profile_folder, file)
                if os.path.isdir(profile_path):
                    continue
                try:
                    # Load Profile and append description
                    temp_profile = openshot.Profile(profile_path)

                    if profile_desc == temp_profile.info.description:
                        profile = self.apply_profile(temp_profile)
                        break
                    if legacy_profile and legacy_profile.Key() == temp_profile.Key():
                        # Switch from legacy profile to new profile
                        legacy_profile_match = temp_profile

                except RuntimeError as e:
                    # This exception occurs when there's a problem parsing the Profile file - display a message and continue
                    log.error("Failed to parse file '%s' as a profile: %s" % (profile_path, e))

        # Upgrade a legacy profile (if needed). Sometimes legacy profiles do match a new profile description, so
        # only upgrade if we have not already found a profile matching the description
        if not profile and legacy_profile_match:
            profile = self.apply_profile(legacy_profile_match)

        return profile

    def apply_profile(self, profile):
        """Apply a specific profile to the current project data"""
        log.info("Setting profile to %s" % profile.info.description)

        # Update default profile
        self._data["profile"] = profile.info.description
        self._data["width"] = profile.info.width
        self._data["height"] = profile.info.height
        self._data["fps"] = {"num": profile.info.fps.num, "den": profile.info.fps.den}
        self._data["display_ratio"] = {"num": profile.info.display_ratio.num, "den": profile.info.display_ratio.den}
        self._data["pixel_ratio"] = {"num": profile.info.pixel_ratio.num, "den": profile.info.pixel_ratio.den}

        # Convert all position, start, and end trims to profile FPS precision (if any)
        change_profile(self._data["clips"] + self._data["effects"] + self._data["markers"], profile)

        return profile

    def load(self, file_path, clear_thumbnails=True):
        """ Load project from file """

        self.new()

        if file_path:
            log.info("Loading project file: %s", file_path)

            # Default project data
            default_project = self._data

            try:
                # Attempt to load v2.X project file
                project_data = self.read_from_file(file_path, path_mode="absolute")

                # Fix history (if broken)
                if not project_data.get("history"):
                    project_data["history"] = {"undo": [], "redo": []}

                # If project has waveforms, enable removing waveforms
                get_app().window.actionClearWaveformData.setEnabled(False)
                for file in project_data["files"]:
                    if file.get("ui",{}).get("audio_data", []):
                        get_app().window.actionClearWaveformData.setEnabled(True)
                        break

            except Exception:
                try:
                    # Attempt to load legacy project file (v1.X version)
                    project_data = self.read_legacy_project_file(file_path)

                except Exception:
                    # Project file not recognized as v1.X or v2.X, bubble up error
                    raise

            # Merge default and project settings, excluding settings not in default.
            self._data = self.merge_settings(default_project, project_data)

            # On success, save current filepath
            self.current_filepath = file_path

            # Update info paths to assets folders
            if clear_thumbnails:
                info.THUMBNAIL_PATH = os.path.join(get_assets_path(self.current_filepath), "thumbnail")
                info.TITLE_PATH = os.path.join(get_assets_path(self.current_filepath), "title")
                info.BLENDER_PATH = os.path.join(get_assets_path(self.current_filepath), "blender")
                info.PROTOBUF_DATA_PATH = os.path.join(get_assets_path(self.current_filepath), "protobuf_data")
                info.CLIPBOARD_PATH = os.path.join(get_assets_path(self.current_filepath), "clipboard")
                info.COMFYUI_OUTPUT_PATH = os.path.join(get_assets_path(self.current_filepath), "comfyui-output")
                info.PROXY_PATH = os.path.join(get_assets_path(self.current_filepath), "optimized")
                migrated = MigrateThumbnailLayout(info.THUMBNAIL_PATH)
                if migrated:
                    log.info("Migrated %s thumbnail(s) to per-file folders", migrated)

            self._migrate_optimized_asset_paths()

            # Clear needs save flag
            self.has_unsaved_changes = False

            # Check if paths are all valid
            self.check_if_paths_are_valid()

            # Clear old thumbnails
            openshot_thumbnails = info.get_default_path("THUMBNAIL_PATH")
            if os.path.exists(openshot_thumbnails) and clear_thumbnails:
                # Clear thumbnails
                shutil.rmtree(openshot_thumbnails, True)
                os.mkdir(openshot_thumbnails)

            # Add to recent files setting
            self.add_to_recent_files(file_path)

            # Upgrade any data structures
            self.upgrade_project_data_structures()

            # Get profile (if any)
            project_profile_desc = self._data.get("profile", "HD 720p 30 fps")
            profile = self.get_profile(profile_desc=project_profile_desc)
            if not profile:
                # Fallback, in case of missing/invalid default profile
                profile = self.get_profile(profile_desc="HD 720p 30 fps")

            # Apply default audio playback settings to this data structure
            self.apply_default_audio_settings()

        # Get app, and distribute all project data through update manager
        get_app().updates.load(self._data)

    def rescale_keyframes(self, scale_factor):
        """Adjust all keyframe coordinates from previous FPS to new FPS (using a scale factor)
           and return scaled project data without modifing the current project."""
        log.info('Scale all keyframes by a factor of %s', scale_factor)
        # Create a scaler instance, and scale all key-frames
        scaler = KeyframeScaler(factor=scale_factor)
        scaler(self._data)

    def read_legacy_project_file(self, file_path):
        """Attempt to read a legacy version 1.x openshot project file"""
        import sys
        import pickle
        from classes.query import File, Track, Clip, Transition
        import openshot
        import json

        # Get translation method
        _ = get_app()._tr

        # Append version info
        project_data = {}
        project_data["version"] = {"openshot-qt": info.VERSION,
                                   "libopenshot": openshot.OPENSHOT_VERSION_FULL}

        # Get FPS from project
        fps = get_app().project.get("fps")
        fps_float = float(fps["num"]) / float(fps["den"])

        # Import legacy openshot classes (from version 1.X)
        from classes.legacy.openshot import classes as legacy_classes
        from classes.legacy.openshot.classes import project as legacy_project
        from classes.legacy.openshot.classes import sequences as legacy_sequences
        from classes.legacy.openshot.classes import track as legacy_track
        from classes.legacy.openshot.classes import clip as legacy_clip
        from classes.legacy.openshot.classes import keyframe as legacy_keyframe
        from classes.legacy.openshot.classes import files as legacy_files
        from classes.legacy.openshot.classes import transition as legacy_transition
        from classes.legacy.openshot.classes import effect as legacy_effect
        from classes.legacy.openshot.classes import marker as legacy_marker
        sys.modules['openshot.classes'] = legacy_classes
        sys.modules['classes.project'] = legacy_project
        sys.modules['classes.sequences'] = legacy_sequences
        sys.modules['classes.track'] = legacy_track
        sys.modules['classes.clip'] = legacy_clip
        sys.modules['classes.keyframe'] = legacy_keyframe
        sys.modules['classes.files'] = legacy_files
        sys.modules['classes.transition'] = legacy_transition
        sys.modules['classes.effect'] = legacy_effect
        sys.modules['classes.marker'] = legacy_marker

        # Keep track of files that failed to load
        failed_files = []

        with open(os.fsencode(file_path), 'rb') as f:
            try:
                # Unpickle legacy openshot project file
                v1_data = pickle.load(f, fix_imports=True, encoding="UTF-8")
                file_lookup = {}

                # Loop through files
                for item in v1_data.project_folder.items:
                    # Is this item a File (i.e. ignore folders)
                    if isinstance(item, legacy_files.OpenShotFile):
                        # Create file
                        try:
                            clip = openshot.Clip(item.name)
                            reader = clip.Reader()
                            file_data = json.loads(reader.Json(), strict=False)

                            # Determine media type
                            file_data["media_type"] = get_media_type(file_data)

                            # Save new file to the project data
                            file = File()
                            file.data = file_data
                            file.save()

                            # Keep track of new ids and old ids
                            file_lookup[item.unique_id] = file

                        except Exception:
                            log.error("%s is not a valid video, audio, or image file",
                                      item.name,
                                      exc_info=1)
                            failed_files.append(item.name)

                # Delete all tracks
                track_list = Track.filter()
                for track in track_list:
                    track.delete()

                # Create new tracks
                track_counter = 0
                for legacy_t in reversed(v1_data.sequences[0].tracks):
                    t = Track()
                    t.data = {"number": track_counter, "y": 0, "label": legacy_t.name}
                    t.save()

                    track_counter += 1

                # Loop through clips
                track_counter = 0
                for sequence in v1_data.sequences:
                    for track in reversed(sequence.tracks):
                        for clip in track.clips:
                            # Get associated file for this clip
                            if clip.file_object.unique_id in file_lookup:
                                file = file_lookup[clip.file_object.unique_id]
                            else:
                                # Skip missing file
                                log.info("Skipping importing missing file: %s" % clip.file_object.unique_id)
                                continue

                            # Create clip
                            if (file.data["media_type"] == "video" or file.data["media_type"] == "image"):
                                # Determine thumb path
                                thumb_path = os.path.join(info.THUMBNAIL_PATH, "%s.png" % file.data["id"])
                            else:
                                # Audio file
                                thumb_path = os.path.join(info.PATH, "images", "AudioThumbnail.png")

                            # Get file name
                            filename = os.path.basename(file.data["path"])

                            file_path = file.absolute_path()

                            # Create clip object for this file
                            c = openshot.Clip(file_path)

                            # Append missing attributes to Clip JSON
                            new_clip = json.loads(c.Json(), strict=False)
                            new_clip["file_id"] = file.id
                            new_clip["title"] = filename

                            # Check for optional start and end attributes
                            new_clip["start"] = clip.start_time
                            new_clip["end"] = clip.end_time
                            new_clip["position"] = clip.position_on_track
                            new_clip["layer"] = track_counter

                            # Clear alpha (if needed)
                            if clip.video_fade_in or clip.video_fade_out:
                                new_clip["alpha"]["Points"] = []

                            # Video Fade IN
                            if clip.video_fade_in:
                                # Add keyframes
                                start = openshot.Point(round(clip.start_time * fps_float) + 1, 0.0, openshot.BEZIER)
                                start_object = json.loads(start.Json(), strict=False)
                                end = openshot.Point(round((clip.start_time + clip.video_fade_in_amount) * fps_float) + 1, 1.0, openshot.BEZIER)
                                end_object = json.loads(end.Json(), strict=False)
                                new_clip["alpha"]["Points"].append(start_object)
                                new_clip["alpha"]["Points"].append(end_object)

                            # Video Fade OUT
                            if clip.video_fade_out:
                                # Add keyframes
                                start = openshot.Point(round((clip.end_time - clip.video_fade_out_amount) * fps_float) + 1, 1.0, openshot.BEZIER)
                                start_object = json.loads(start.Json(), strict=False)
                                end = openshot.Point(round(clip.end_time * fps_float) + 1, 0.0, openshot.BEZIER)
                                end_object = json.loads(end.Json(), strict=False)
                                new_clip["alpha"]["Points"].append(start_object)
                                new_clip["alpha"]["Points"].append(end_object)

                            # Clear Audio (if needed)
                            if clip.audio_fade_in or clip.audio_fade_out:
                                new_clip["volume"]["Points"] = []
                            else:
                                p = openshot.Point(1, clip.volume / 100.0, openshot.BEZIER)
                                p_object = json.loads(p.Json(), strict=False)
                                new_clip["volume"] = {"Points": [p_object]}

                            # Audio Fade IN
                            if clip.audio_fade_in:
                                # Add keyframes
                                start = openshot.Point(round(clip.start_time * fps_float) + 1, 0.0, openshot.BEZIER)
                                start_object = json.loads(start.Json(), strict=False)
                                end = openshot.Point(round((clip.start_time + clip.video_fade_in_amount) * fps_float) + 1, clip.volume / 100.0, openshot.BEZIER)
                                end_object = json.loads(end.Json(), strict=False)
                                new_clip["volume"]["Points"].append(start_object)
                                new_clip["volume"]["Points"].append(end_object)

                            # Audio Fade OUT
                            if clip.audio_fade_out:
                                # Add keyframes
                                start = openshot.Point(round((clip.end_time - clip.video_fade_out_amount) * fps_float) + 1, clip.volume / 100.0, openshot.BEZIER)
                                start_object = json.loads(start.Json(), strict=False)
                                end = openshot.Point(round(clip.end_time * fps_float) + 1, 0.0, openshot.BEZIER)
                                end_object = json.loads(end.Json(), strict=False)
                                new_clip["volume"]["Points"].append(start_object)
                                new_clip["volume"]["Points"].append(end_object)

                            # Save clip
                            clip_object = Clip()
                            clip_object.data = new_clip
                            clip_object.save()

                        # Loop through transitions
                        for trans in track.transitions:
                            # Fix default transition
                            if not trans.resource or not os.path.exists(trans.resource):
                                trans.resource = os.path.join(info.PATH, "transitions", "common", "fade.svg")

                            # Open up QtImageReader for transition Image
                            transition_reader = openshot.QtImageReader(trans.resource)

                            trans_begin_value = 1.0
                            trans_end_value = -1.0
                            if trans.reverse:
                                trans_begin_value = -1.0
                                trans_end_value = 1.0

                            brightness = openshot.Keyframe()
                            brightness.AddPoint(1, trans_begin_value, openshot.BEZIER)
                            brightness.AddPoint(round(trans.length * fps_float) + 1, trans_end_value, openshot.BEZIER)
                            contrast = openshot.Keyframe(trans.softness * 10.0)

                            # Create transition dictionary
                            transitions_data = {
                                "id": get_app().project.generate_id(),
                                "layer": track_counter,
                                "title": "Transition",
                                "type": "Mask",
                                "position": trans.position_on_track,
                                "start": 0,
                                "end": trans.length,
                                "brightness": json.loads(brightness.Json(), strict=False),
                                "contrast": json.loads(contrast.Json(), strict=False),
                                "reader": json.loads(transition_reader.Json(), strict=False),
                                "replace_image": False
                            }

                            # Save transition
                            t = Transition()
                            t.data = transitions_data
                            t.save()

                        # Increment track counter
                        track_counter += 1

            except Exception as ex:
                # Error parsing legacy contents
                msg = "Failed to load legacy project file %(path)s" % {"path": file_path}
                log.error(msg, exc_info=1)
                raise RuntimeError(msg) from ex

        # Show warning if some files failed to load
        if failed_files:
            # Throw exception
            raise RuntimeError("Failed to load the following files:\n%s" % ", ".join(failed_files))

        # Return mostly empty project_data dict (with just the current version #)
        log.info("Successfully loaded legacy project file: %s", file_path)
        return project_data

    def upgrade_project_data_structures(self):
        """Fix any issues with old project files (if any)"""
        openshot_version = self._data["version"]["openshot-qt"]
        libopenshot_version = self._data["version"]["libopenshot"]

        log.info("Project data: openshot %s, libopenshot %s",
                 openshot_version, libopenshot_version)

        if openshot_version == "0.0.0":
            # If version = 0.0.0, this is the beta of OpenShot
            # Fix alpha values (they are now flipped)
            for clip in self._data["clips"]:
                # Loop through keyframes for alpha
                for point in clip["alpha"]["Points"]:
                    # Flip the alpha value
                    if "co" in point:
                        point["co"]["Y"] = 1.0 - point["co"]["Y"]
                    if "handle_left" in point:
                        point["handle_left"]["Y"] = 1.0 - point["handle_left"]["Y"]
                    if "handle_right" in point:
                        point["handle_right"]["Y"] = 1.0 - point["handle_right"]["Y"]

        elif openshot_version <= "2.1.0-dev":
            # Fix handle_left and handle_right coordinates and default to ease in/out bezier curve
            # using the new percent based keyframes
            for clip_type in ["clips", "effects"]:
                for clip in self._data[clip_type]:
                    for object in [clip] + clip.get('effects', []):
                        for item_key, item_data in object.items():
                            # Does clip attribute have a {"Points": [...]} list
                            if type(item_data) == dict and "Points" in item_data:
                                for point in item_data.get("Points"):
                                    # Convert to percent-based curves
                                    if "handle_left" in point:
                                        # Left handle
                                        point.get("handle_left")["X"] = 0.5
                                        point.get("handle_left")["Y"] = 1.0
                                    if "handle_right" in point:
                                        # Right handle
                                        point.get("handle_right")["X"] = 0.5
                                        point.get("handle_right")["Y"] = 0.0

                            elif type(item_data) == dict and "red" in item_data:
                                for color in ["red", "blue", "green", "alpha"]:
                                    for point in item_data.get(color).get("Points"):
                                        # Convert to percent-based curves
                                        if "handle_left" in point:
                                            # Left handle
                                            point.get("handle_left")["X"] = 0.5
                                            point.get("handle_left")["Y"] = 1.0
                                        if "handle_right" in point:
                                            # Right handle
                                            point.get("handle_right")["X"] = 0.5
                                            point.get("handle_right")["Y"] = 0.0

        elif openshot_version.startswith("2.5."):
            # Replace missing crop properties from 2.5.x
            log.debug("Scanning OpenShot 2.5 project for legacy cropping")
            for clip in self._data.get("clips", []):
                # Loop through keyframes for alpha
                crop_x = clip.pop("crop_x", {})
                crop_y = clip.pop("crop_y", {})
                crop_width = clip.pop("crop_width", {})
                crop_height = clip.pop("crop_height", {})

                if any([self.is_keyframe_valid(crop_x, 0.0),
                        self.is_keyframe_valid(crop_y, 0.0),
                        self.is_keyframe_valid(crop_width, 1.0),
                        self.is_keyframe_valid(crop_height, 1.0),
                       ]):
                    # Apply crop effect to clip (which wasn't available in 2.5.x)
                    log.info("Migrating OpenShot 2.5 crop properties for clip %s", clip.get("id", "<unknown>"))
                    from json import loads as jl
                    effect = openshot.EffectInfo().CreateEffect("Crop")
                    effect.Id(get_app().project.generate_id())
                    effect_json = jl(effect.Json())

                    # Attach previous clip crop values
                    effect_json.update({
                        "x": crop_x or jl(openshot.Keyframe(0.0).Json()),
                        "y": crop_y or jl(openshot.Keyframe(0.0).Json()),
                        "right": crop_width or jl(openshot.Keyframe(1.0).Json()),
                        "bottom": crop_height or jl(openshot.Keyframe(1.0).Json()),
                    })

                    # Reverse values on right & bottom
                    for prop in ["right", "bottom"]:
                        for point in effect_json[prop].get("Points", []):
                            point["co"]["Y"] = 1.0 - point.get("co", {}).get("Y", 0.0)

                    # Append effect JSON to clip
                    clip["effects"].append(effect_json)


        elif openshot_version <= "3.1.1":

            # Reverse alpha value for Tracker & Object Detector tracked boxes
            log.debug("Scanning OpenShot project for legacy TrackedObjectBBox (background_alpha, stroke_alpha)")

            for clip in self._data.get("clips", []):
                for effect in clip.get("effects", []):
                    if effect.get("name") in ["Tracker", "Object Detector"]:

                        # Update display_box_text
                        if "display_box_text" in effect:
                            log.info("Migrating legacy Object Detector display_box_text property "
                                     "for clip %s", clip.get("id", "<unknown>"))
                            display_box_text_points = effect.get("display_box_text", {}).get("Points", [])
                            for point in display_box_text_points:
                                if "co" in point:
                                    display_box_text = point.get("co", {}).get("Y", 1.0)
                                    point["co"]["Y"] = 1.0 - display_box_text
                            if not display_box_text_points:
                                # Default to text visible
                                display_box_text_points.append(json.loads(openshot.Point(1.0).Json()))

                        # Update tracked objects
                        objects = effect.get("objects", {})
                        for tracked_key, tracked_data in objects.items():
                            log.info("Migrating legacy TrackedObjectBBox alpha properties "
                                     "for clip %s and tracked object: %s", clip.get("id", "<unknown>"), tracked_key)

                            # Update Child Clip ID to Parent property (if any)
                            child_clip_id = tracked_data.get("child_clip_id")
                            if child_clip_id:
                                for child_clip in self._data.get("clips", []):
                                    if child_clip.get("id") == child_clip_id:
                                        log.info(f"Migrating child_clip_id {child_clip_id} to parent property for tracked object {tracked_key}")
                                        child_clip["parentObjectId"] = tracked_key

                            # Update background_alpha points directly
                            background_alpha_points = tracked_data.get("background_alpha", {}).get("Points", [])
                            for point in background_alpha_points:
                                if "co" in point:
                                    background_alpha = point.get("co", {}).get("Y", 1.0)
                                    point["co"]["Y"] = 1.0 - background_alpha

                            # Update stroke_alpha points directly
                            stroke_alpha_points = tracked_data.get("stroke_alpha", {}).get("Points", [])
                            for point in stroke_alpha_points:
                                if "co" in point:
                                    stroke_alpha = point.get("co", {}).get("Y", 1.0)
                                    point["co"]["Y"] = 1.0 - stroke_alpha

        # Fix default project id (if found)
        if self._data.get("id") == "T0":
            self._data["id"] = self.generate_id()

    def is_keyframe_valid(self, keyframe, default_value):
        """Check if a keyframe is not empty (i.e. > 1 point, or a non default_value)"""
        points = keyframe.get("Points", [])
        if not points or not isinstance(points, list):
            return False
        return any([
            len(points) > 1,
            points[0].get("co", {}).get("Y", default_value) != default_value,
        ])

    def save(self, file_path, backup_only=False):
        """ Save project file to disk """
        import openshot

        log.info("Saving project file: %s", file_path)

        # Move all temp files (i.e. Blender Animations, Titles, Thumbnails, Protobuf files) to the project folder
        if not backup_only:
            self.move_temp_paths_to_project_folder(
                file_path, previous_path=self.current_filepath)

        # Append version info
        self._data["version"] = {"openshot-qt": info.VERSION,
                                 "libopenshot": openshot.OPENSHOT_VERSION_FULL}

        # Try to save project settings file, will raise error on failure
        self.write_to_file(
            file_path,
            self._data,
            path_mode="ignore" if backup_only else "relative",
            previous_path=self.current_filepath if not backup_only else None)

        if not backup_only:
            # On success, save current filepath
            self.current_filepath = file_path

            # Update info paths to assets folders
            info.THUMBNAIL_PATH = os.path.join(get_assets_path(self.current_filepath), "thumbnail")
            info.TITLE_PATH = os.path.join(get_assets_path(self.current_filepath), "title")
            info.BLENDER_PATH = os.path.join(get_assets_path(self.current_filepath), "blender")
            info.CLIPBOARD_PATH = os.path.join(get_assets_path(self.current_filepath), "clipboard")
            info.COMFYUI_OUTPUT_PATH = os.path.join(get_assets_path(self.current_filepath), "comfyui-output")
            info.PROXY_PATH = os.path.join(get_assets_path(self.current_filepath), "optimized")

            self.add_to_recent_files(file_path)
            self.has_unsaved_changes = False

    @staticmethod
    def _paths_match(path_a, path_b):
        """Return True when both paths resolve to the same local filesystem path."""
        if not path_a or not path_b:
            return False
        return os.path.normcase(os.path.abspath(path_a)) == os.path.normcase(os.path.abspath(path_b))

    def _should_move_runtime_assets(self, source_root, default_name):
        """Return True when a source root is the active runtime temp directory."""
        default_root = info.get_default_path(default_name)
        return self._paths_match(source_root, default_root)

    def _sync_asset_entry(self, source_path, target_path, move=False, replace_existing=False):
        """Copy or move a file/folder into the target asset tree."""
        if not os.path.exists(source_path):
            return

        if os.path.isdir(source_path):
            if os.path.exists(target_path) and not os.path.isdir(target_path):
                os.remove(target_path)

            if replace_existing and os.path.isdir(target_path):
                shutil.rmtree(target_path, True)

            if not os.path.exists(target_path):
                parent_path = os.path.dirname(target_path)
                if parent_path:
                    os.makedirs(parent_path, exist_ok=True)
                if move:
                    shutil.move(source_path, target_path)
                else:
                    shutil.copytree(source_path, target_path)
                return

            for child_name in os.listdir(source_path):
                self._sync_asset_entry(
                    os.path.join(source_path, child_name),
                    os.path.join(target_path, child_name),
                    move=move,
                    replace_existing=replace_existing,
                )

            if move and os.path.isdir(source_path) and not os.listdir(source_path):
                os.rmdir(source_path)
            return

        if os.path.exists(target_path):
            if replace_existing:
                if os.path.isdir(target_path):
                    shutil.rmtree(target_path, True)
                else:
                    os.remove(target_path)
            elif move:
                os.remove(source_path)
                return
            else:
                return

        parent_path = os.path.dirname(target_path)
        if parent_path:
            os.makedirs(parent_path, exist_ok=True)
        if move:
            shutil.move(source_path, target_path)
        else:
            shutil.copy2(source_path, target_path)

    def _sync_asset_root(self, source_root, target_root, move=False, replace_existing=False):
        """Copy or move all entries from a source asset root into a target root."""
        if not source_root or not os.path.exists(source_root):
            return
        if self._paths_match(source_root, target_root):
            return

        os.makedirs(target_root, exist_ok=True)
        for entry_name in os.listdir(source_root):
            self._sync_asset_entry(
                os.path.join(source_root, entry_name),
                os.path.join(target_root, entry_name),
                move=move,
                replace_existing=replace_existing,
            )

    def move_temp_paths_to_project_folder(self, file_path, previous_path=None):
        """ Move all temp files (such as Thumbnails, Titles, and Blender animations) to the project asset folder. """
        try:
            # Get or generate asset folder name, max 30 chars of filename + "_assets"
            asset_path = get_assets_path(file_path)
            target_thumb_path = os.path.join(asset_path, "thumbnail")
            target_title_path = os.path.join(asset_path, "title")
            target_blender_path = os.path.join(asset_path, "blender")
            target_protobuf_path = os.path.join(asset_path, "protobuf_data")
            target_clipboard_path = os.path.join(asset_path, "clipboard")
            target_comfy_output_path = os.path.join(asset_path, "comfyui-output")
            target_proxy_path = os.path.join(asset_path, "optimized")

            # Create any missing target paths
            try:
                for target_dir in [asset_path, target_thumb_path, target_title_path,
                                   target_blender_path, target_protobuf_path,
                                   target_clipboard_path, target_comfy_output_path,
                                   target_proxy_path]:
                    if not os.path.exists(target_dir):
                        os.mkdir(target_dir)
            except OSError:
                pass

            # Update paths (if a previous path exists)
            #   /Project1/ to /Project2/ for example
            if previous_path:
                previous_asset_path = get_assets_path(previous_path)
                info.THUMBNAIL_PATH = os.path.join(previous_asset_path, "thumbnail")
                info.TITLE_PATH = os.path.join(previous_asset_path, "title")
                info.BLENDER_PATH = os.path.join(previous_asset_path, "blender")
                info.PROTOBUF_DATA_PATH = os.path.join(previous_asset_path, "protobuf_data")
                info.CLIPBOARD_PATH = os.path.join(previous_asset_path, "clipboard")
                info.COMFYUI_OUTPUT_PATH = os.path.join(previous_asset_path, "comfyui-output")
                info.PROXY_PATH = os.path.join(previous_asset_path, "optimized")

            # Track assets we copy/update
            copied_assets = {
                "blender": set(),
                "title": set(),
                "clipboard": set(),
                "comfyui_output": set(),
                "proxy": set(),
            }
            reader_paths = {}

            self._sync_asset_root(
                info.THUMBNAIL_PATH,
                target_thumb_path,
                move=self._should_move_runtime_assets(info.THUMBNAIL_PATH, "THUMBNAIL_PATH"),
            )
            self._sync_asset_root(
                info.TITLE_PATH,
                target_title_path,
                move=self._should_move_runtime_assets(info.TITLE_PATH, "TITLE_PATH"),
            )
            self._sync_asset_root(
                info.BLENDER_PATH,
                target_blender_path,
                move=self._should_move_runtime_assets(info.BLENDER_PATH, "BLENDER_PATH"),
            )
            self._sync_asset_root(
                info.CLIPBOARD_PATH,
                target_clipboard_path,
                move=self._should_move_runtime_assets(info.CLIPBOARD_PATH, "CLIPBOARD_PATH"),
            )
            self._sync_asset_root(
                info.COMFYUI_OUTPUT_PATH,
                target_comfy_output_path,
                move=self._should_move_runtime_assets(info.COMFYUI_OUTPUT_PATH, "COMFYUI_OUTPUT_PATH"),
                replace_existing=True,
            )
            self._sync_asset_root(
                info.PROTOBUF_DATA_PATH,
                target_protobuf_path,
                move=self._should_move_runtime_assets(info.PROTOBUF_DATA_PATH, "PROTOBUF_DATA_PATH"),
            )
            self._sync_asset_root(
                info.PROXY_PATH,
                target_proxy_path,
                move=self._should_move_runtime_assets(info.PROXY_PATH, "PROXY_PATH"),
            )

            # Copy any necessary assets for File records
            for file in self._data["files"]:
                path = file["path"]
                file_id = file["id"]

                # For now, store thumbnail path for backwards compatibility
                file["image"] = os.path.join(target_thumb_path, f"{file_id}.png")

                # Assets which need to be copied
                new_asset_path = None
                if info.BLENDER_PATH in path:
                    path_abs = os.path.abspath(path)
                    blender_root_abs = os.path.abspath(info.BLENDER_PATH)
                    if path_abs.startswith(blender_root_abs + os.sep):
                        relative_blender_path = os.path.relpath(path_abs, blender_root_abs)
                        blender_root_name = relative_blender_path.split(os.sep, 1)[0]
                        if blender_root_name and blender_root_name not in copied_assets["blender"]:
                            source_blender_dir = os.path.join(info.BLENDER_PATH, blender_root_name)
                            if os.path.abspath(source_blender_dir) != os.path.abspath(target_blender_path):
                                copied_assets["blender"].add(blender_root_name)
                                log.info("Copied dir %s to %s", blender_root_name, target_blender_path)
                        new_asset_path = os.path.join(target_blender_path, relative_blender_path)

                if info.TITLE_PATH in path:
                    # Copy title files into assets folder
                    old_dir, asset_name = os.path.split(path)
                    if asset_name not in copied_assets["title"]:
                        # Copy title into assets title folder
                        if os.path.abspath(old_dir) != os.path.abspath(target_title_path):
                            copied_assets["title"].add(asset_name)
                            log.info("Copied title %s to %s", asset_name, target_title_path)
                    new_asset_path = os.path.join(target_title_path, asset_name)

                if info.CLIPBOARD_PATH in path:
                    old_dir, asset_name = os.path.split(path)
                    if asset_name not in copied_assets["clipboard"]:
                        if os.path.abspath(old_dir) != os.path.abspath(target_clipboard_path):
                            copied_assets["clipboard"].add(asset_name)
                            log.info("Copied clipboard %s to %s", asset_name, target_clipboard_path)
                    new_asset_path = os.path.join(target_clipboard_path, asset_name)

                comfy_output_abs = os.path.abspath(info.COMFYUI_OUTPUT_PATH)
                path_abs = os.path.abspath(path)
                if path_abs.startswith(comfy_output_abs + os.sep):
                    if os.path.abspath(os.path.dirname(path)) != os.path.abspath(target_comfy_output_path):
                        relative_output_path = os.path.relpath(path_abs, comfy_output_abs)
                        if relative_output_path not in copied_assets["comfyui_output"]:
                            copied_assets["comfyui_output"].add(relative_output_path)
                            log.info("Copied ComfyUI output %s to %s", relative_output_path, target_comfy_output_path)
                        new_asset_path = os.path.join(target_comfy_output_path, relative_output_path)

                # Update path in File object to new location
                if new_asset_path:
                    file["path"] = new_asset_path
                    reader_paths[file_id] = new_asset_path
                    log.info("Set file %s path to %s", file_id, new_asset_path)

                proxy_reader = file.get("proxy_reader")
                proxy_path = proxy_reader.get("path") if isinstance(proxy_reader, dict) else ""
                if proxy_path and info.PROXY_PATH in proxy_path:
                    proxy_dir, proxy_name = os.path.split(proxy_path)
                    if proxy_name not in copied_assets["proxy"]:
                        if os.path.abspath(proxy_dir) != os.path.abspath(target_proxy_path):
                            copied_assets["proxy"].add(proxy_name)
                            source_proxy = os.path.join(proxy_dir, proxy_name)
                            target_proxy = os.path.join(target_proxy_path, proxy_name)
                            if os.path.isdir(source_proxy):
                                if os.path.exists(target_proxy):
                                    shutil.rmtree(target_proxy, True)
                                shutil.copytree(source_proxy, target_proxy)
                            elif os.path.exists(source_proxy):
                                shutil.copy2(source_proxy, target_proxy)
                            log.info("Copied proxy %s to %s", proxy_name, target_proxy_path)
                    proxy_reader["path"] = os.path.join(target_proxy_path, proxy_name)

            # Copy all Clip thumbnails and update reader paths
            for clip in self._data["clips"]:
                file_id = clip["file_id"]
                clip_id = clip["id"]

                # For now, store thumbnail path for backwards compatibility
                clip["image"] = os.path.join(target_thumb_path, f"{file_id}.png")

                log.info("Checking clip %s path for file %s", clip_id, file_id)
                # Update paths to files stored in our working space or old path structure
                # (should have already been copied during previous File stage)
                if file_id and file_id in reader_paths:
                    clip["reader"]["path"] = reader_paths[file_id]
                    log.info("Updated clip %s path for file %s", clip_id, file_id)

                log.info("Checking effects in clip %s path for protobuf files" % clip_id)
                for effect in clip.get("effects", []):
                    if "protobuf_data_path" in effect:
                        old_protobuf_path = effect["protobuf_data_path"]
                        old_protobuf_dir, protobuf_name = os.path.split(old_protobuf_path)
                        if old_protobuf_dir != target_protobuf_path:
                            effect["protobuf_data_path"] = os.path.join(target_protobuf_path, protobuf_name)
                            log.info("Copied protobuf %s to %s", old_protobuf_path, target_protobuf_path)

        except Exception:
            log.error(
                "Error while moving temp paths to project assets folder %s",
                asset_path, exc_info=1)

    def add_to_recent_files(self, file_path):
        """ Add this project to the recent files list """
        if not file_path or file_path is info.BACKUP_FILE:
            # Ignore backup recovery project
            return

        s = get_app().get_settings()
        recent_projects = s.get("recent_projects")

        normalized_path = normalized_local_path(file_path)
        normalized_key = comparable_local_path(normalized_path)

        # Remove existing project entries, including mixed-separator duplicates.
        recent_projects = [
            normalized_local_path(existing_path)
            for existing_path in recent_projects
            if comparable_local_path(existing_path) != normalized_key
        ]

        # Remove oldest item (if needed)
        if len(recent_projects) >= 10:
            del recent_projects[0]

        # Append file path to end of recent files
        recent_projects.append(normalized_path)

        # Save setting
        s.set("recent_projects", recent_projects)
        s.save()

    def check_if_paths_are_valid(self):
        """Check if all paths are valid, and prompt to update them if needed"""
        app = get_app()
        settings = app.get_settings()
        # Get translation method
        _ = app._tr

        log.info("checking project files...")
        prompt_state = {"cancelled": False, "missing_path_decisions": {}}

        def _path_is_missing(path):
            return bool(path) and "%" not in path and not os.path.exists(path)

        def _resolve_missing_path(path):
            """Prompt to locate a missing path, returning resolved path or empty when skipped."""
            if not _path_is_missing(path):
                return path, False

            # Reuse prior decision for this exact missing path in this load pass.
            # This avoids duplicate prompts when the same asset path appears in
            # multiple places (e.g. top-level effect + clip effect).
            missing_path_decisions = prompt_state.setdefault("missing_path_decisions", {})
            if path in missing_path_decisions:
                cached_path = missing_path_decisions[path]
                if cached_path:
                    return cached_path, False
                return "", True

            found_path, is_modified, is_skipped = find_missing_file(path, prompt_state)
            if found_path and is_modified and not is_skipped:
                settings.setDefaultPath(settings.actionType.IMPORT, found_path)
                missing_path_decisions[path] = found_path
                return found_path, False
            if is_skipped:
                missing_path_decisions[path] = ""
                skip_mode = prompt_state.get("last_skip")
                if skip_mode == "all":
                    skip_detail = " (user selected Skip All)"
                else:
                    skip_detail = " (user selected Skip File)"
                # Use warning level so this is visible even when info logs are filtered.
                log.warning(
                    "Missing path skipped during project load%s: %s",
                    skip_detail,
                    path,
                )
            return "", True

        def _collect_effect_path_refs(effect):
            """Collect mutable refs to path-like fields used by effects."""
            refs = []
            seen = set()

            self._drop_obsolete_effect_resource(effect)

            def _add_ref(container, key):
                if not isinstance(container, dict):
                    return
                value = container.get(key, "")
                if not isinstance(value, str) or not value:
                    return
                ref_id = (id(container), key)
                if ref_id in seen:
                    return
                seen.add(ref_id)
                refs.append((container, key, value))

            def _walk(obj, parent_key=""):
                if isinstance(obj, dict):
                    for key, value in obj.items():
                        if isinstance(value, dict):
                            if isinstance(value.get("path"), str) and "reader" in key.lower():
                                _add_ref(value, "path")
                            _walk(value, key)
                        elif isinstance(value, list):
                            _walk(value, key)
                        elif isinstance(value, str):
                            if key in {"resource", "protobuf_data_path", "lut_path", "image"}:
                                _add_ref(obj, key)
                            elif key == "path" and "reader" in parent_key.lower():
                                _add_ref(obj, key)
                elif isinstance(obj, list):
                    for item in obj:
                        _walk(item, parent_key)

            _walk(effect)
            return refs

        def _first_missing_effect_path(effect):
            for _, _, path in _collect_effect_path_refs(effect):
                if _path_is_missing(path):
                    return path
            return ""

        def _repair_effect_paths(effect):
            """Prompt for missing paths on an effect-like dict."""
            if not isinstance(effect, dict):
                return False

            path_refs = _collect_effect_path_refs(effect)
            missing_paths = []
            for _, _, path in path_refs:
                if _path_is_missing(path) and path not in missing_paths:
                    missing_paths.append(path)

            for missing_path in missing_paths:
                resolved_path, should_remove = _resolve_missing_path(missing_path)
                if should_remove:
                    log.warning(
                        "Removing effect with unresolved path. effect_id=%s missing_path=%s",
                        effect.get("id", ""),
                        missing_path,
                    )
                    return False
                for container, key, current_path in path_refs:
                    if current_path == missing_path:
                        container[key] = resolved_path

            return True

        # Loop through each files in natural-sorted filename order
        files_sorted = sorted(
            self._data["files"],
            key=lambda f: os.path.basename(f.get("path", "")).lower()
        )
        for file in files_sorted:
            path = file["path"]
            parent_path, file_name_with_ext = os.path.split(path)

            log.info("checking file %s", path)
            if not os.path.exists(path) and "%" not in path:
                # File is missing - prompt/resolve using the shared path-decision cache.
                resolved_path, should_remove = _resolve_missing_path(path)
                if should_remove:
                    log.info('Removed missing file: %s', file_name_with_ext)
                    self._data["files"].remove(file)
                elif resolved_path and resolved_path != path:
                    file["path"] = resolved_path
                    settings.setDefaultPath(settings.actionType.IMPORT, resolved_path)
                    log.info("Auto-updated missing file: %s", resolved_path)

        # Build a lookup of valid file IDs after any removals/updates
        file_paths_by_id = {file.get("id"): file.get("path") for file in self._data["files"]}

        # Loop through each clip (in reverse order). Don't prompt for clips.
        for clip in reversed(self._data["clips"]):
            file_id = clip.get("file_id") or clip.get("reader", {}).get("id")
            if file_id and file_id in file_paths_by_id:
                # Keep clip reader path in sync with its File path
                reader = clip.get("reader")
                if not isinstance(reader, dict):
                    reader = {}
                    clip["reader"] = reader
                reader["path"] = file_paths_by_id[file_id]

            path = clip.get("reader", {}).get("path", "")

            if file_id and file_id not in file_paths_by_id:
                # Remove clips with invalid/missing files
                file_name_with_ext = os.path.basename(path) if path else ""
                log.info('Removed missing clip: %s', file_name_with_ext)
                self._data["clips"].remove(clip)
                continue

            if path and not os.path.exists(path) and "%" not in path:
                # Remove missing clip without prompting
                file_name_with_ext = os.path.basename(path)
                log.info('Removed missing clip: %s', file_name_with_ext)
                self._data["clips"].remove(clip)

        # Loop through top-level effects/transitions and prompt for missing reader/resource paths.
        for effect in reversed(self._data["effects"]):
            self._drop_obsolete_effect_resource(effect)
            if not _repair_effect_paths(effect):
                missing_path = _first_missing_effect_path(effect)
                effect_name = os.path.basename(missing_path) if missing_path else effect.get("id", "")
                log.info("Removed missing effect: %s", effect_name)
                self._data["effects"].remove(effect)

        # Some clip effects can also carry reader/resource paths (e.g. mask/image-driven effects).
        for clip in self._data["clips"]:
            effects = clip.get("effects")
            if not isinstance(effects, list):
                continue
            for effect in reversed(effects):
                self._drop_obsolete_effect_resource(effect)
                if not _repair_effect_paths(effect):
                    missing_path = _first_missing_effect_path(effect)
                    effect_name = os.path.basename(missing_path) if missing_path else effect.get("id", "")
                    log.info("Removed missing clip effect on %s: %s", clip.get("id", ""), effect_name)
                    effects.remove(effect)

    def changed(self, action):
        """ This method is invoked by the UpdateManager each time a change happens (i.e UpdateInterface) """
        updates = get_app().updates

        def mark_dirty():
            if not self.has_unsaved_changes:
                log.debug(
                    "Project dirty flag set: action=%s key=%s ignore_history=%s values=%s",
                    action.type,
                    action.key,
                    updates.ignore_history,
                    action.values,
                )
            self.has_unsaved_changes = True

        if action.type == "insert":
            # Insert new item
            old_vals = self._set(action.key, action.values, add=True)
            action.set_old_values(old_vals)  # Save previous values to reverse this action
            mark_dirty()

        elif action.type == "update":
            # Update existing item
            old_vals = self._set(action.key, action.values)
            action.set_old_values(old_vals)  # Save previous values to reverse this action
            mark_dirty()

            if len(action.key) == 1 and action.key[0] in ["fps"]:
                # FPS changed (apply profile)
                profile_key = self._data.get("profile")
                profile = self.get_profile(profile_key)
                if profile:
                    # Get current FPS (prior to changing)
                    new_fps = self._data.get("fps")
                    new_fps_float = float(new_fps["num"]) / float(new_fps["den"])
                    old_fps_float = float(old_vals["num"]) / float(old_vals["den"])
                    fps_factor = float(new_fps_float / old_fps_float)

                    if fps_factor != 1.0:
                        log.info(f"Convert {old_fps_float} FPS to {new_fps_float} FPS (profile: {profile.ShortName()})")
                        # Snap to new FPS grid (start, end, duration)
                        change_profile(self._data["clips"] + self._data["effects"], profile)

                        # Rescale keyframes to match new FPS
                        self.rescale_keyframes(fps_factor)

                    # Update size of audio-only files
                    for file in self._data.get("files", []):
                        # Check for audio-only files
                        if file.get("has_audio") and not file.get("has_video"):
                            # Audio-only file should match the current project size and FPS
                            file["width"] = profile.info.width
                            file["height"] = profile.info.height
                            file["display_ratio"]["num"] = profile.info.display_ratio.num
                            file["display_ratio"]["den"] = profile.info.display_ratio.den

                            # Change all related clips
                            for clip in self._data.get("clips", []):
                                if clip.get("reader", {}).get("id") == file.get("id"):
                                    clip["reader"] = file

                    # Broadcast this change out
                    get_app().updates.load(self._data, reset_history=False)
                else:
                    log.warning(f"No profile found for {profile_key}")

        elif action.type == "delete":
            # Delete existing item
            old_vals = self._set(action.key, remove=True)
            action.set_old_values(old_vals)  # Save previous values to reverse this action
            mark_dirty()

        elif action.type == "load":
            # Don't track unsaved changes when loading a project
            pass

    # Utility methods
    def generate_id(self, digits=10):
        """ Generate random alphanumeric ids """

        chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
        id = ""
        for i in range(digits):
            c_index = random.randint(0, len(chars) - 1)
            id += (chars[c_index])
        return id

    def apply_default_audio_settings(self):
        """Apply the default preferences for sampleRate and channels to
        the current project data, to force playback at a specific rate and for
        a specific # of audio channels and channel layout."""
        s = get_app().get_settings()

        # Get the default audio settings for the timeline (and preview playback)
        default_sample_rate = int(s.get("default-samplerate"))
        default_channel_layout = s.get("default-channellayout")

        channels = 2
        channel_layout = openshot.LAYOUT_STEREO
        if default_channel_layout == "LAYOUT_MONO":
            channels = 1
            channel_layout = openshot.LAYOUT_MONO
        elif default_channel_layout == "LAYOUT_STEREO":
            channels = 2
            channel_layout = openshot.LAYOUT_STEREO
        elif default_channel_layout == "LAYOUT_SURROUND":
            channels = 3
            channel_layout = openshot.LAYOUT_SURROUND
        elif default_channel_layout == "LAYOUT_5POINT1":
            channels = 6
            channel_layout = openshot.LAYOUT_5POINT1
        elif default_channel_layout == "LAYOUT_7POINT1":
            channels = 8
            channel_layout = openshot.LAYOUT_7POINT1

        # Set default samplerate and channels
        self._data["sample_rate"] = default_sample_rate
        self._data["channels"] = channels
        self._data["channel_layout"] = channel_layout

        log.info("Apply default audio playback settings: %s, %s channels" % (self._data["sample_rate"], self._data["channels"]))



"""
 @file
 @brief Helpers for resolving media paths to absolute/relative forms
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2025 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
"""

import html
import os
import re

from classes import info
from classes.app import get_app
from classes.assets import get_assets_path


def _project_file_path(project_file=None):
    """Return the active project file path (if any)."""
    if project_file:
        return project_file
    app = get_app()
    if app and hasattr(app, "project"):
        return getattr(app.project, "current_filepath", None)
    return None


def _project_folder(project_file=None):
    project_file = _project_file_path(project_file)
    if project_file:
        return os.path.dirname(project_file)
    return info.HOME_PATH


def _token_suffix(path_value):
    parts = path_value.split("/", 1)
    if len(parts) == 2:
        return parts[1]
    return ""


def absolute_media_path(path_value, project_file=None):
    """Resolve OpenShot-specific tokens and relative paths into absolute paths."""
    if not path_value:
        return ""

    normalized = path_value.replace("\\", "/")

    if normalized.startswith("@emojis"):
        suffix = _token_suffix(normalized)
        return os.path.normpath(os.path.join(info.PATH, "emojis", "color", "svg", suffix))

    if normalized.startswith("@transitions"):
        suffix = _token_suffix(normalized)
        return os.path.normpath(os.path.join(info.PATH, "transitions", suffix))

    if normalized.startswith("@colors"):
        suffix = _token_suffix(normalized)
        return os.path.normpath(os.path.join(info.COLORS_PATH, suffix))

    if normalized.startswith("@assets"):
        project_file = _project_file_path(project_file)
        assets_root = get_assets_path(project_file, create_paths=False)
        suffix = _token_suffix(normalized)
        return os.path.normpath(os.path.join(assets_root, suffix))

    if normalized.startswith("thumbnail/"):
        project_file = _project_file_path(project_file)
        assets_root = get_assets_path(project_file, create_paths=False)
        return os.path.normpath(os.path.join(assets_root, normalized.replace("thumbnail/", "thumbnail" + os.sep)))

    if os.path.isabs(normalized):
        return os.path.normpath(normalized)

    base_folder = _project_folder(project_file)
    return os.path.normpath(os.path.join(base_folder, normalized))


def relative_export_path(abs_path, export_folder):
    """Return path relative to export folder when possible."""
    if not abs_path:
        return ""
    try:
        abs_norm = os.path.normpath(abs_path)
        if not export_folder:
            return abs_norm.replace("\\", "/")
        export_norm = os.path.normpath(export_folder)
        if os.name == "nt":
            src_drive = os.path.splitdrive(abs_norm)[0].lower()
            dst_drive = os.path.splitdrive(export_norm)[0].lower()
            if src_drive and dst_drive and src_drive != dst_drive:
                return abs_norm.replace("\\", "/")
        rel_path = os.path.relpath(abs_norm, export_norm)
        return rel_path.replace("\\", "/")
    except Exception:
        return abs_path.replace("\\", "/")


def absolute_path_from_export(path_value, base_folder, project_file=None):
    """Resolve a relative path stored in an export back into an absolute path."""
    if not path_value:
        return ""

    normalized = path_value.replace("\\", "/")

    if normalized.startswith("@"):
        return absolute_media_path(normalized, project_file)

    if os.path.isabs(normalized):
        return os.path.normpath(normalized)

    if not base_folder:
        base_folder = _project_folder(project_file)

    return os.path.normpath(os.path.join(base_folder, normalized))


def normalize_path(path_value):
    """Return a path string with POSIX separators (useful for XML)."""
    if not path_value:
        return ""
    return path_value.replace("\\", "/")


def normalized_local_path(path_value):
    """Return a normalized local filesystem path for storage/display."""
    if not path_value:
        return ""
    return os.path.normpath(os.path.abspath(path_value))


def comparable_local_path(path_value):
    """Return a normalized local path suitable for equality checks."""
    normalized = normalized_local_path(path_value)
    if not normalized:
        return ""
    return os.path.normcase(normalized)


def native_display_path(path_value):
    """Return a display path using the current platform's separators."""
    return normalized_local_path(path_value)


def wrapped_path_html(path_value):
    """Return HTML for a path that wraps cleanly after path separators."""
    display_path = native_display_path(path_value)
    if not display_path:
        return ""
    escaped_path = html.escape(display_path)
    return re.sub(r"([/\\\\])", r"\1<wbr/>", escaped_path)


def comparable_media_path(path_value, project_file=None):
    """Return a normalized absolute media path suitable for equality checks."""
    resolved = absolute_media_path(path_value, project_file)
    if not resolved:
        return ""
    return os.path.normcase(os.path.normpath(resolved))


def media_paths_equal(path_a, path_b, project_file=None):
    """Compare two media paths after token expansion and platform normalization."""
    if not path_a or not path_b:
        return False
    return comparable_media_path(path_a, project_file) == comparable_media_path(path_b, project_file)



"""
 @file
 @brief This file loads and saves settings
 @author Noah Figg <eggmunkee@hotmail.com>
 @author Jonathan Thomas <jonathan@openshot.org>
 @author Olivier Girard <eolinwen@gmail.com>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

# SettingStore - class which allows getting/storing of settings, loading and saving to json
import os
from enum import Enum, unique, auto

from classes import info
from classes.app import get_app
from classes.logger import log
from classes.json_data import JsonDataStore
import openshot


class SettingStore(JsonDataStore):
    """ This class only allows setting pre-existing keys taken from default settings file, and merges user settings
    on load, assumes default OS dir."""

    @unique
    class pathType(Enum):
        RECENT=auto() # Return the last used path for the same action
        PROJECT=auto() # Return the current project path

    @unique
    class actionType(Enum):
        IMPORT=auto()
        EXPORT=auto()
        LOAD=auto()
        SAVE=auto()

    def __init__(self, parent=None):
        super().__init__()
        # Also keep track of our parent, if defined
        self.app = parent
        # Set the data type name for logging clarity (base class functions use this variable)
        self.data_type = "user settings"
        self.settings_filename = "openshot.settings"
        self.defaults_path = os.path.join(info.PATH, 'settings', '_default.settings')

    def _apply_runtime_defaults(self, settings_list):
        """Overlay runtime-detected libopenshot defaults onto selected settings."""
        lib_settings = openshot.Settings.Instance()
        runtime_defaults = {
            "omp_threads_number": lib_settings.DefaultOMPThreads(),
            "ff_threads_number": lib_settings.DefaultFFThreads(),
        }

        for item in settings_list:
            setting_name = item.get("setting")
            if setting_name in runtime_defaults:
                item["value"] = runtime_defaults[setting_name]

        return settings_list

    def has_user_value(self, key):
        """Return True when the user settings file contains an explicit value for key."""
        key = key.lower()
        file_path = os.path.join(info.USER_PATH, self.settings_filename)
        if not os.path.exists(os.fsencode(file_path)):
            return False

        try:
            user_settings = self.read_from_file(file_path)
        except Exception:
            return False

        for item in user_settings:
            if item.get("setting", "").lower() == key and "value" in item:
                return True
        return False

    def get_all_settings(self):
        """ Get the entire list of settings (with all metadata) """
        return self._data

    def set(self, key, value):
        """ Store setting, but adding isn't allowed. All possible settings must be in default settings file. """
        key = key.lower()

        # Index settings values (for easy updates)
        user_values = {
            item["setting"].lower(): item
            for item in self._data
            if all(["setting" in item, "value" in item])
        }

        # Save setting
        if key in user_values:
            user_values[key].update({"value": value})
        else:
            log.warn(
                "{} key '{}' not valid. The following are valid: {}".format(
                    self.data_type,
                    key,
                    list(self._data.keys()),
                ))

    def load(self):
        """ Load user settings file from disk, merging with allowed settings in default settings file.
        Creates user settings if missing. """

        # try to load default settings, on failure will raise exception to caller
        default_settings = self._apply_runtime_defaults(
            self.read_from_file(self.defaults_path)
        )
        self._data = default_settings

        # Try to find user settings dir, give up if it's not there
        if not os.path.exists(info.USER_PATH):
            return True

        # Load or create user settings
        file_path = os.path.join(info.USER_PATH, self.settings_filename)
        if os.path.exists(os.fsencode(file_path)):
            try:
                user_settings = self.read_from_file(file_path)
                # Merge sources, excluding user settings not found in default
                self._data = self.merge_settings(default_settings, user_settings)
            except Exception as ex:
                log.error("Error loading settings file: %s", ex)
                if self.app:
                    # We have a parent, ask to show a message box
                    self.app.settings_load_error(file_path)

        # Return success of saving user settings file back after merge
        return self.write_to_file(file_path, self._data)

    def save(self):
        """ Save user settings file to disk """

        # Try to find user settings file
        if os.path.exists(info.USER_PATH):
            file_path = os.path.join(info.USER_PATH, self.settings_filename)
            # try to save data to file, will raise exception on failure
            self.write_to_file(file_path, self._data)

    def restore(self, category_filter=None):
        """
        Restore settings to default, optionally filtering by category, and preserving specific keys.
        Return True if any settings with 'restart: True' are changed.
        """
        log.info(f"Restoring defaults for category: {category_filter or 'all categories'}")
        preserve_keys = ['unique_install_id', 'tutorial_ids', 'tutorial_enabled', 'send_metrics', 'recent_projects']

        requires_restart = False  # Track if any setting requires a restart

        try:
            default_settings = self._apply_runtime_defaults(
                self.read_from_file(self.defaults_path)
            )
        except Exception as ex:
            log.error(f"Error loading default settings: {ex}")
            return False

        if os.path.exists(info.USER_PATH):
            file_path = os.path.join(info.USER_PATH, self.settings_filename)
            try:
                user_settings = self.read_from_file(file_path) if os.path.exists(os.fsencode(file_path)) else []

                # Convert user settings to a dict for easier lookup
                user_settings_dict = {item['setting']: item for item in user_settings}

                # Keep settings outside the filter or in the preserved keys
                updated_settings = [
                    item for item in user_settings
                    if item.get("category") != category_filter or item.get("setting") in preserve_keys
                ]

                # Add default settings for the current category, ignoring preserved keys
                for item in default_settings:
                    if item.get("category") == category_filter and item.get("setting") not in preserve_keys:
                        default_value = item.get("value")
                        user_value = user_settings_dict.get(item['setting'], {}).get("value")

                        # If the value has changed, and restart is required, set the flag
                        if item.get("restart") == True and default_value != user_value:
                            requires_restart = True

                        # Append the default setting to updated settings
                        updated_settings.append(item)

                # Save the updated settings back to the user file
                self.write_to_file(file_path, updated_settings)
                log.info(f"Settings updated successfully for category: {category_filter}")

            except Exception as ex:
                log.error(f"Error saving settings: {ex}")
                return False

        self._data = updated_settings
        return requires_restart

    def pathSettings(self, action: actionType):
        """Given an action, return the corresponding setting names"""
        if action == self.actionType.IMPORT:
            return {'type':"locationImportType", 'path':"locationImportPath"}
        if action == self.actionType.EXPORT:
            return {'type': "locationExportType", 'path': "locationExportPath"}
        if action in [self.actionType.SAVE, self.actionType.LOAD]:
            return {'type': "locationProjectType", 'path': "locationProjectPath"}
        log.info("Action not recognized")
        log.error("Given action is not valid %s" % action)
        return None

    def setDefaultPath(self, action: actionType, recent_path: str):
        """
        Change the path setting corresponding to the given action
        """
        if not os.path.isdir(recent_path):
            log.info(f"reducing file: {recent_path} to dir: {os.path.dirname(recent_path)}")
            recent_path = os.path.dirname(recent_path)
        if not os.path.exists(recent_path):
            log.error(f"{recent_path} is not a valid path")
            return
        if info.USER_PATH == recent_path:
            log.info(f"Ignore setting recent path: {recent_path}, most likely due to backup recovery")
            return

        try:
            setting = self.pathSettings(action)
            log.debug(f"Setting {setting['path']} to {recent_path}")
            self.set(setting["path"], recent_path)
        except:
            log.error(f"Error; action: {action}; recent_path: {recent_path}")

    def getDefaultPath(self, action: actionType):
        """
        Returns the starting path for file browsing
        - validates paths before returning them
        - Actions: import, export, project open/save-as
        - Types
            RECENT: return the last directory where this action occurred
            PROJECT: return the directory of the current project
        """
        _ = self.app._tr

        default_path = ""
        project_path = self.app.project.current_filepath or ""
        setting = self.pathSettings(action)

        try:
            recentOrProject = self.get(setting["type"])
        except:
            log.error(f"getting setting %s" % setting["type"])

        if recentOrProject == self.pathType.PROJECT.value:
            log.debug("Using Project Path")
            default_path = project_path
        else:
            try:
                log.debug("Using Recent Path")
                default_path = self.get(setting["path"])
            except:
                log.error("getting setting %s" % setting["path"])

        if os.path.isfile(default_path):
            # Make sure default_path is a directory
            default_path = os.path.dirname(default_path)

        if not (default_path and os.path.exists(default_path)):
            log.debug("Default path invalid. Falling back to home directory")
            return os.path.expanduser("~")

        return default_path
